/******************************************************************************
 *
 * nicstar.c   (NICStAR2 driver)
 *
 * Date: 12/09/2011
 * Version: 2.39
 *
 * LINUX Device driver supporting UBR, CBR, VBR and ABR for PROATM cards 
 *
 *
 * IMPORTANT:   This code is derived from the Linux nicstar driver 
 *              from Rui Prior. Large portions of this driver still exist here.
 *
 * Author: Christian Bucari: bucari@prosum.net
 * Copyright (c) 2002 - 2011 Prosum 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. 
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************
 * Note on VBR usage:
 *
 * VBR QoS is characterized by pcr, scr and mbs parameters.
 * So far, the atm_trafprm structure of linux/atm.h does not include neither 
 * scr nor mbs. We could provide a patch to modify atm_traf_prm, however we 
 * found simpler to reuse existing parameters. This makes the job without 
 * impacting other ATM layers and drivers.
 * So when using VBR VC's, enter the pcr value into max_pcr, the scr value into
 * pcr and the the mbs value into min_pcr.
 * This VBR implementation is intended mostly for writing ATM  applications 
 * that manage the qos structure directly. Original Linux-atm tools do not 
 * provide VBR support but Prosum provides a slightly modified linux-atm 
 * package supporting the VBR QoS. 
 *
 */  

/* Header files ***************************************************************/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/atmdev.h>
#include <linux/atm.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/bitops.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include "nicstar.h"
#include "suni.h"
#include "nicstar_rtbl.c"

/* Configurable parameters ****************************************************/

#undef GENERAL_DEBUG
#undef EXTRA_DEBUG

#undef PHY_LOOPBACK  
 
#undef NS_USE_DESTRUCTORS /* For now keep this undefined unless you know
                                you're going to use only raw ATM */
#undef NS_POLLING

/* Do not touch these *********************************************************/

#ifdef GENERAL_DEBUG
#define PRINTK(args...) printk(args)
#else
#define PRINTK(args...)
#endif /* GENERAL_DEBUG */

#ifdef EXTRA_DEBUG
#define XPRINTK(args...) printk(args)
#else 
#define XPRINTK(args...)
#endif /* EXTRA_DEBUG */

/* Macros *********************************************************************/

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

#define CMD_BUSY(card) (readl((card)->membase + STAT) & NS_STAT_CMDBZ)

#define NS_DELAY mdelay(1)

#define ALIGN_BUS_ADDR(addr, alignment) \
        ((((u32) (addr)) + (((u32) (alignment)) - 1)) & ~(((u32) (alignment)) - 1))

#define ALIGN_ADDRESS(addr, alignment) \
        ((((unsigned long) (addr)) + (((unsigned long) (alignment)) - 1)) \
                                        & ~(((unsigned long) (alignment)) - 1))

#if LINUX_VERSION_CODE < KERNEL_VERSION (2,6,14)
#define TIMESTP(sb)  do_gettimeofday(&(sb)->stamp)
#define SKB_UNLINK(a,b) skb_unlink((a))
#else
#define TIMESTP(sb)  __net_timestamp((sb))
#define SKB_UNLINK(a,b) skb_unlink((a),(b))
#endif

#undef CEIL

#define ns_grab_int_lock(card,flags)   spin_lock_irqsave(&(card)->int_lock,(flags))
#define ns_grab_res_lock(card,flags)   spin_lock_irqsave(&(card)->res_lock,(flags))
#define ns_grab_scq_lock(card,scq,flags)  spin_lock_irqsave(&(scq)->lock,flags)

#define LED_ON \
do { \
    volatile  u32 gp = readl(card->membase + GP) & 0xfffffff8;  \
    udelay(5);  \
    writel (gp, card->membase + GP);  \
    udelay(5); \
} while(0)

#define LED_OFF \
do { \
    volatile  u32 gp = readl(card->membase + GP) & 0xfffffff8;  \
    udelay(5);  \
    writel (gp | NS_GP_EEDO, card->membase + GP);  \
    udelay(5); \
} while(0)

/* Function prototypes ******************************************************/
static u32 ns_read_sram(ns_dev *card, u32 sram_address);
static void ns_write_sram(ns_dev *card, u32 sram_address, u32 *value, int count);
static int __devinit ns_init_card(int i, struct pci_dev *pcidev);
static void __devinit ns_init_card_error(ns_dev *card, int error);
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,19)
static irqreturn_t ns_irq_handler(int irq, void *dev_id);
static int ns_open(struct atm_vcc *vcc);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static irqreturn_t ns_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static int ns_open(struct atm_vcc *vcc);
#else
static void ns_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static int ns_open(struct atm_vcc *vcc, short vpi, int vci);
#endif
static void  ns_close(struct atm_vcc *vcc);
static scq_info *get_scq(ns_dev *card, int type, u32 scd);
static void free_scq(ns_dev *card, scq_info *scq, struct atm_vcc *vcc);
static void push_rxbufs(ns_dev *card, char type, struct sk_buff *skb);
static void fill_tst(ns_dev *card, int n, vc_map *vc, u32 conn);
static int ns_send(struct atm_vcc *vcc, struct sk_buff *skb);
static int push_scqe(ns_dev *card, vc_map *vc, scq_info *scq, ns_scqe *tbd, 
                          struct sk_buff *skb);
static void process_tsq(ns_dev *card);
static void drain_scq(ns_dev *card, scq_info *scq, vc_map *vc, int pos);
static void process_rsq(ns_dev *card);
static void dequeue_rx(ns_dev *card, ns_rsqe *rsqe);

#ifdef NS_USE_DESTRUCTORS
static void ns_sb_destructor(struct sk_buff *sb);
static void ns_lb_destructor(struct sk_buff *lb);
static void ns_hb_destructor(struct sk_buff *hb);

#endif /* NS_USE_DESTRUCTORS */
static void recycle_rx_buf(ns_dev *card, struct sk_buff *skb);
static void recycle_iovec_rx_bufs(ns_dev *card, struct iovec *iov, int count);
static void recycle_iov_buf(ns_dev *card, struct sk_buff *iovb);
static void dequeue_sm_buf(ns_dev *card, struct sk_buff *sb);
static void dequeue_lg_buf(ns_dev *card, struct sk_buff *lb);
static int ns_proc_read(struct atm_dev *dev, loff_t *pos, char *page);
static int ns_ioctl(struct atm_dev *dev, unsigned int cmd, void *arg);
static void which_list(ns_dev *card, struct sk_buff *skb);
static void ns_eeprom_rd (ns_dev *card, u32 ee_add, u8 *memptr, u32 len);
#ifdef NS_POLLING
static void ns_poll(unsigned long arg);
#endif 
static int ns_parse_mac(char *mac, unsigned char *esi);
static short ns_h2i(char c);
static void ns_phy_put(struct atm_dev *dev, unsigned char value, 
                           unsigned long addr);
static unsigned char ns_phy_get(struct atm_dev *dev, unsigned long addr);
static inline void waitfor_notbusy(struct ns_dev *card);
static void ns_push_on_scq (ns_dev *card, scq_info *scq);
static int ns_phys_init (ns_dev *card, u32 *);
static inline int ns_find_conn (ns_dev *card, short vpi, int vci);
static int ns_check_vc (ns_dev *card, short vpi, int vci);

#ifdef NS_RCQ_SUPPORT
static int ns_process_rcqe (ns_dev *card, ns_rcqe *rawcell);
static void ns_rawc(ns_dev *card);
#endif

#ifdef NS_OAM_SUPPORT
static int ns_oam_type (u32 pti, u32 vci);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
static void ns_push_oamcell (struct work_struct *);
#else
static void ns_push_oamcell (void *dev_id);
#endif
static void ns_push_on_oam_tq (ns_dev *card, ns_rcqe *rawcell);
static void ns_activate_deactivate_cc(ns_dev *card, ns_rcqe *rawcell);
static int ns_process_oam_cell (ns_dev *card, ns_rcqe *rawcell);

static u8 ns_all1[16] = {0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
                         0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};
static u8 ns_all0[16] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                         0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
static u8 ns_llid[16] = NS_LLID;
#endif 

/* Global variables ***********************************************************/
static struct ns_dev *cards[NS_MAX_CARDS];
static unsigned num_cards = 0;

static struct atmdev_ops atm_ops = 
{
   .open            = ns_open,
   .close           = ns_close,
   .ioctl           = ns_ioctl,
   .send            = ns_send,
   .phy_put         = ns_phy_put,
   .phy_get         = ns_phy_get,
   .proc_read       = ns_proc_read,
   .owner           = THIS_MODULE,
};

#ifdef NS_POLLING
static struct timer_list ns_timer;
#endif
static char *mac[NS_MAX_CARDS] = {
    NULL
#if NS_MAX_CARDS > 1
    , NULL
#endif
#if NS_MAX_CARDS > 2
    , NULL
#endif
#if NS_MAX_CARDS > 3
    , NULL
#endif
#if NS_MAX_CARDS > 4
    , NULL
#endif
}
;
static int vpibits = NS_VPIBITS;
static int vcibase = NS_VCIBASE;
static int vpibase = NS_VPIBASE;
#ifdef NS_SUNI_FULL_SDH
static int fullsdh = 1;
#else
static int fullsdh = 0;
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,0)
module_param_array(mac, charp, NULL, 0);
MODULE_PARM_DESC(mac, "mac address in the form XX:XX:XX:XX:XX:XX");
module_param(vpibits, int, 0);
MODULE_PARM_DESC(vpibits, "0, 1, 2 or 8");
module_param(vcibase, int, 0);
MODULE_PARM_DESC(vcibase, "see readme file");
module_param(vpibase, int, 0);
MODULE_PARM_DESC(vpibase, "see readme file");
module_param(fullsdh, int, 0);
MODULE_PARM_DESC(fullsdh, "see readme file");
#else
#ifdef MODULE
MODULE_PARM(mac, "1-"__MODULE_STRING(NS_MAX_CARDS)"s");
MODULE_PARM_DESC(mac, "mac address in the form XX:XX:XX:XX:XX:XX");
MODULE_PARM(vpibits, "i");
MODULE_PARM_DESC(vpibits, "0, 1, 2 or 8");
MODULE_PARM(vcibase, "i");
MODULE_PARM_DESC(vcibase, "see readme file");
MODULE_PARM(vpibase, "i");
MODULE_PARM_DESC(vpibase, "see readme file");
MODULE_PARM(fullsdh, "i");
MODULE_PARM_DESC(fullsdh, "see readme file");
#endif 
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
 #define nfmark mark
 #define NS_INIT_WORK(a,b,c) INIT_WORK((a),(b))
#else
 #define NS_INIT_WORK(a,b,c) INIT_WORK((a),(b),(c))
#endif

 

/* Functions *******************************************************************/

#ifdef NET_SKBUFF_DATA_USES_OFFSET
 static inline unsigned char *skb_end_ptr(const struct sk_buff *skb)
 {
         return skb->head + skb->end;
 }
#else
 static inline unsigned char *skb_end_ptr(const struct sk_buff *skb)
 {
         return skb->end;
 }
 #endif

static int __devinit nicstar_init_one(struct pci_dev *pcidev,
                                      const struct pci_device_id *ent)
{
   static int index = 0;

   PRINTK("nicstar%d: nicstar_init_one()\n", index);

   cards[index] = NULL;
   if (ns_init_card (index, pcidev)) {
      cards[index] = NULL;      /* don't increment index */
      return -ENODEV;
   }
   index++;
   return 0;
}

static void __devexit nicstar_remove_one(struct pci_dev *pcidev)
{
   int i, j;
   ns_dev *card = pci_get_drvdata(pcidev);
   struct sk_buff *hb;
   struct sk_buff *iovb;
   struct sk_buff *lb;
   struct sk_buff *sb;
   
   PRINTK("nicstar: nicstar_remove_one()\n");

   i = card->index;

   if (cards[i] == NULL)
      return;

   /* reset SAR */
   writel(0, card->membase + CFG);

   if (card->atmdev) {
       atm_dev_deregister(card->atmdev);

       /* stop phy */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,22)
  #ifdef CONFIG_ATM_NICSTAR_USE_SUNI
            if (card->max_pcr == ATM_OC3_PCR) {
                (*card->atmdev->phy->stop)(card->atmdev);
            }
  #endif
#else
       if (card->atmdev->phy && card->atmdev->phy->stop)
          (*card->atmdev->phy->stop)(card->atmdev);
#endif
   }

   /* Disable PCI device */
   pci_disable_device(pcidev);

   /* Free up resources */
   j = 0;
   PRINTK("nicstar%d: freeing %d huge buffers.\n", i, card->hbpool.count);
   while ((hb = skb_dequeue(&card->hbpool.queue)) != NULL) {
      dev_kfree_skb_any(hb);
      j++;
   }
   PRINTK("nicstar%d: %d huge buffers freed.\n", i, j);
   j = 0;
   PRINTK("nicstar%d: freeing %d iovec buffers.\n", i, card->iovpool.count);
   while ((iovb = skb_dequeue(&card->iovpool.queue)) != NULL){
      dev_kfree_skb_any(iovb);
      j++;
   }
   PRINTK("nicstar%d: %d iovec buffers freed.\n", i, j);
   while ((lb = skb_dequeue(&card->bpool[1].queue)) != NULL)
      dev_kfree_skb_any(lb);

   while ((sb = skb_dequeue(&card->bpool[0].queue)) != NULL)
      dev_kfree_skb_any(sb);

   PRINTK("nicstar%d: freeing SCQ0.\n",i);
   free_scq(card, card->scq0, NULL);

   PRINTK("nicstar%d: freeing SCQs.\n", i);
   for (j = 0; j < NS_SCD_NUM; j++) {
        if (card->scd2vc[j] != NULL)
            free_scq(card, card->scd2vc[j]->scq, card->scd2vc[j]->tx_vcc);
   }
#ifdef NS_OAM_SUPPORT
   PRINTK("nicstar%d: freeing OAM skb queue.\n", i);
   while ((sb = skb_dequeue (&card->oam_tq)))    
        dev_kfree_skb_any (sb);
#endif

   PRINTK("nicstar%d: freeing card.\n", i);
   free_pages((unsigned long)card->vcmap, get_order(card->rct_size *sizeof(vc_map)));
   pci_free_consistent(card->pcidev, NS_RSQSIZE + NS_RSQ_ALIGNMENT, 
                                            card->rsq.org, card->rsq.org_pa);
   pci_free_consistent(card->pcidev, NS_TSQSIZE + NS_TSQ_ALIGNMENT, 
                                            card->tsq.org, card->tsq.org_pa);
   pci_free_consistent(card->pcidev, 2 * sizeof(u32), 
                                            card->raw_hnd, card->raw_hnd_pa);
   free_irq(card->pcidev->irq, card);
   iounmap((void *) card->membase);
   kfree(card);

}

static struct pci_device_id nicstar_pci_tbl[] __devinitdata =
{
        { PCI_VENDOR_ID_IDT, PCI_DEVICE_ID_IDT_IDT77252,
          PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
        { 0, }
};
MODULE_DEVICE_TABLE(pci, nicstar_pci_tbl);

static struct pci_driver nicstar_driver = {
        .name           = "nicstar",
        .id_table       = nicstar_pci_tbl,
        .probe          = nicstar_init_one,
        .remove         = __devexit_p(nicstar_remove_one),
};

static int __init nicstar_init(void)
{
    unsigned error = 0;  /* Initialized to remove compile warning */
        struct pci_dev *pci_bridge = NULL;
        u8 cache_size;

    /*
         * On some architectures, the default cache line size is not supported by
         * the bridge, which reduces the performance. Try to set a supported value.
         */
        pci_bridge = pci_get_device(PCI_VENDOR_ID_PLX, 0x8112, NULL);
    while (pci_bridge) {
        if (pci_read_config_byte(pci_bridge, PCI_CACHE_LINE_SIZE, &cache_size)) {
                    printk("nicstar: error: cannot read bridge cache line size\n"); 
                    pci_dev_put(pci_bridge);
                    return -EINVAL;
            }
            if (cache_size > 32) {
                    cache_size = 32;
                    if (pci_write_config_byte(pci_bridge, PCI_CACHE_LINE_SIZE, cache_size))
                            printk("nicstar: cannot set cache line size to %d\n", cache_size);
                        else
                            printk("nicstar: set the bridge cache line size to %d\n", cache_size);
            }
            pci_dev_put(pci_bridge);
        pci_bridge = pci_get_device(PCI_VENDOR_ID_PLX, 0x8112, pci_bridge);
        }

    /* 
     * Print some compilation options 
     */
    XPRINTK("nicstar: Extra debug enabled.\n");
    PRINTK("nicstar: General debug enabled.\n");

#ifdef NS_RCQ_SUPPORT
    printk ("nicstar: version %s, support of raw cells enabled\n", NICSTAR2_VERSION);
#else
    printk ("nicstar: version %s, support of raw cells disabled\n", NICSTAR2_VERSION);
#endif

#if defined(NS_RCQ_SUPPORT) && defined(NS_AAL0_TO_RCQ)
    printk ("nicstar: Cells received on AAL0 VCs are routed to the raw cell queue\n");
#else
    printk ("nicstar: Cells received on AAL0 VCs cells are routed to standard queues (need EOM bit=1)\n"); 
#endif

#if defined(NS_RCQ_SUPPORT) && defined (NS_OAM_SUPPORT)
    printk ("nicstar: Support of OAM cells enabled\n");
#else
    printk ("nicstar: Support of OAM cells disabled\n");
#endif


   PRINTK("nicstar: nicstar_init()\n");





#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 22)
   error = pci_register_driver (&nicstar_driver); 
#else
   error = pci_module_init (&nicstar_driver);
#endif
   if (error)
        return error;

#ifdef PHY_LOOPBACK
   printk("nicstar: using PHY loopback.\n");
#endif /* PHY_LOOPBACK */

#ifdef NS_POLLING 
   init_timer(&ns_timer);
   ns_timer.expires = jiffies + NS_POLL_PERIOD;
   ns_timer.data = 0UL;
   ns_timer.function = ns_poll;
   add_timer(&ns_timer);
#endif

   return 0;
}

static void __exit nicstar_cleanup(void)
{
   PRINTK("nicstar: nicstar_cleanup()\n");

#ifdef NS_POLLING 
   del_timer(&ns_timer);
#endif
   pci_unregister_driver(&nicstar_driver);

}

static inline void waitfor_notbusy(struct ns_dev *card)
{
    u32             stat;

    stat = readl(card->membase + STAT);
    while (stat & NS_STAT_CMDBZ)
        stat = readl((card)->membase + STAT);
}

/* these functions use 32-bit word address */

static u32 ns_read_sram(ns_dev *card, u32 sram_address)
{
    unsigned long   flags;
    u32             data;
    sram_address    <<= 2;
    sram_address &= 0x001FFFFC;                       /* 2 MB max */ 
    sram_address |= 0x50000000;                       /* SRAM read command */
    ns_grab_res_lock(card, flags);
    waitfor_notbusy(card);
    writel(sram_address, card->membase + CMD);
    waitfor_notbusy(card);
    data = readl(card->membase + DR0);
    spin_unlock_irqrestore(&card->res_lock, flags);
    return data;
}

static void ns_write_sram(ns_dev *card, u32 sram_address, u32 *value, int count)
{
    unsigned long   flags;
    int             i, c;

    count--;                        /* count range now is 0..3 instead of 1..4 */
    c = count;
    c <<= 2;                        /* to use increments of 4 */
    ns_grab_res_lock(card, flags);
    waitfor_notbusy(card);
    for (i = 0; i <= c; i += 4)
       writel(* (value++), card->membase + i);

    /* Note: DR# registers are the first 4 dwords in nicstar's memspace,
                so card->membase + DR0 = card->membase */
    sram_address <<= 2;
    sram_address &= 0x001FFFFC;                   /* 2 MB max */
    sram_address |= (0x40000000 | count);
    writel(sram_address, card->membase + CMD);
    spin_unlock_irqrestore(&card->res_lock, flags);
    waitfor_notbusy(card);
}

static int __devinit ns_sram_check (ns_dev *card) 
{
    u32 pattern[] = {0x00000000, 0xFFFFFFFF, 0x55555555, 0xAAAAAAAA};
    int bits = card->sram_size==512? 19: card->sram_size==128? 17: 15;
    int i;
    u32 addr, data;

    /* check SRAM data bus */
    for (i=0; i < sizeof(pattern)/sizeof(u32); i++) {
        ns_write_sram(card, 0x1000, pattern+i, 1);
        if ((data = ns_read_sram(card, 0x1000)) != pattern[i]) {
            printk ("nicstar%d: SRAM error: write 0x%08X, read 0x%08X\n", 
                                    card->index, pattern[i], data);
            return -1;
        }
    }

    /* check SRAM address bus */
    for (i=0, addr=1; i<bits; i++, addr <<=1) {
        ns_write_sram(card, addr, pattern+1, 1);
        ns_write_sram(card, 0, pattern, 1);
        if ((data = ns_read_sram(card, addr)) != pattern[1]) {
            printk ("nicstar%d: SRAM error: address bit %d stuck to GND\n", 
                                    card->index, i);
            return -1;
        }
        ns_write_sram(card, addr, pattern, 1);

        ns_write_sram(card, ~addr, pattern+1, 1);
        ns_write_sram(card, 0xffffffff, pattern, 1);
        if ((data = ns_read_sram(card, ~addr)) != pattern[1]) {
            printk ("nicstar%d: SRAM error: address bit %d stuck to VCC\n", 
                                    card->index, i);
            return -1;
        }
        ns_write_sram(card, ~addr, pattern, 1); 
    }
    return 0;

}

static int __devinit ns_phys_init (ns_dev *card, u32 *ns_cfg)
{
    u32 data;

    PRINTK("nicstar%d: ns_phys_init ()\n", card->index);

    /* PHY reset */
    writel(0xa, card->membase + GP);
    NS_DELAY;
    writel(0x2, card->membase + GP);
    NS_DELAY;
    waitfor_notbusy(card);
    XPRINTK("nicstar%d: DR0 read (0x%08X).\n", card->index, 
                                        readl(card->membase + DR0));
    NS_DELAY;

    /* Detect PHY type */
    printk("nicstar%d: PHY type detection called.\n", card->index);
    waitfor_notbusy(card);
    writel(NS_CMD_READ_UTILITY | 0x00000100, card->membase + CMD);
    waitfor_notbusy(card);
    data = readl(card->membase + DR0);
    XPRINTK("nicstar%d: DR0 read (0x%08X).\n", card->index, data);
    strcpy(card->name, "unknown_card");

    switch (data) {
#ifdef CONFIG_ATM_NICSTAR_USE_SUNI
      case 0x00000030:
      case 0x00000031:
        printk("nicstar%d: PHY seems to be 155 Mbps.\n", card->index);
        strcpy(card->name, "proatm155");
        card->max_pcr = ATM_OC3_PCR;
#ifdef PHY_LOOPBACK
        waitfor_notbusy(card);
        writel(0x00000002, card->membase + DR0);
        writel(NS_CMD_WRITE_UTILITY | 0x00000105, card->membase + CMD);
#endif /* PHY_LOOPBACK */
        break;
#endif /* USE_SUNI */
      default:
        printk("nicstar%d: unknown PHY type (0x%08X).\n", card->index, data);
        return 8;
    }
    writel(0x00000000, card->membase + GP);

    /* Find SRAM size */
    data = 0x11111111;
    ns_write_sram(card, 0x4000, &data, 1);
    data = 0x22222222;
    ns_write_sram(card, 0x14000, &data, 1);
    data = 0x33333333;
    ns_write_sram(card, 0x54000, &data, 1);

    XPRINTK ("SRAM(0x4000) = 0x%x\n", ns_read_sram(card, 0x4000)); 
    XPRINTK ("SRAM(0x14000) = 0x%x\n", ns_read_sram(card, 0x14000)); 
    XPRINTK ("SRAM(0x54000) = 0x%x\n", ns_read_sram(card, 0x54000)); 

    if (ns_read_sram(card, 0x4000) == 0x11111111) {
        if (ns_read_sram(card, 0x14000) == 0x22222222) {
            card->sram_size = 512;   /* 512K words */
            card->rct_size = 16384;
            card->vcibits = 14 - vpibits;
            card->tst_num = 4095;
            *ns_cfg |= NS_CFG_RCTSIZE_16384_ENTRIES;
        }
        else {
            card->sram_size = 128;     /* 128K words */
            card->rct_size = 4096;
            card->vcibits = 12 - vpibits;
            card->tst_num = 4095;
            *ns_cfg |= NS_CFG_RCTSIZE_4096_ENTRIES;
        }
    }
    else {
        card->sram_size = 32;      /* 32K words */
        card->rct_size = 1024;
        card->vcibits = 10 - vpibits;
        card->tst_num = 2047;
        *ns_cfg |= NS_CFG_RCTSIZE_1024_ENTRIES;
    }
    printk("nicstar%d: SRAM size = %dK x 32bit\n", card->index, card->sram_size);

    if (ns_sram_check(card))       /* cannot hurt */
        return 8;

    card->vpibits = vpibits;
    card->maxvpi = (1<<card->vpibits) - 1;
    PRINTK ("    VPI bits = %d, VPI mask = 0x%x\n", card->vpibits, card->maxvpi);
    card->maxvci = (1<<card->vcibits) - 1;
    PRINTK ("    VCI bits = %d, VCI mask = 0x%x\n", card->vcibits, card->maxvci);

    /* dynamic SRAM map */
    card->rct = NS_TCT_ENTRY_SIZE * card->rct_size;
    card->rate = card->rct + NS_RCT_ENTRY_SIZE * card->rct_size + 2*NS_RFBQ_SIZE;
    card->tst =  card->rate + abr_vbr_rate_tables_len;
    card->dtst = card->tst + card->tst_num + 1;
    card->scd = card->dtst + NS_DTST_SIZE ;
    card->rxfifo = card->sram_size * 1024 - NS_RXFIFO_SIZE;
    card->scd_size = ((card->rxfifo - card->scd)/NS_SCD_ENTRY_SIZE) - 1;
    if (card->scd_size >= 16383)
        card->scd_size = 16383;
    card->scd_ubr0 = card->scd + card->scd_size * NS_SCD_ENTRY_SIZE;

    XPRINTK ("nicstar%d: SRAM Map for %d connections:\n", 
                                        card->index, card->rct_size);
    XPRINTK ("    TCT =      0x0\n");
    XPRINTK ("    RCT =      0x%x\n", card->rct);
    XPRINTK ("    RATE_TBL = 0x%x\n", card->rate);
    XPRINTK ("    TST =      0x%x\n", card->tst);
    XPRINTK ("    DTST =     0x%x\n", card->dtst);
    XPRINTK ("    SCD =      0x%x\n", card->scd);
    XPRINTK ("    SCD_UBR0 = 0x%x\n", card->scd_ubr0);
    XPRINTK ("    RXFIFO =   0x%x\n", card->rxfifo);
    XPRINTK ("    SCD number = %d\n", card->scd_size);

    return 0;
}

static int __devinit ns_init_card(int i, struct pci_dev *pcidev)
{
    int             j;
    struct ns_dev   *card = NULL;
    unsigned char   pci_latency;
    unsigned        error = 0;
    u32             data;
    u32             u32d[4];
    u32             ns_cfg = 0;
    u32             *abr_vbr_rate_tables;
    u32             bcount;
    u32             cfg[9] = { NS_CFG_VPIBITS_0, NS_CFG_VPIBITS_1, NS_CFG_VPIBITS_2, 
                                            0,0,0,0,0, NS_CFG_VPIBITS_8}; 
    u32             membase;

    PRINTK("nicstar%d: ns_init_card()\n", i);

    if (pci_enable_device(pcidev)) {
        printk("nicstar%d: can't enable PCI device\n", i);
        error = 2;
        ns_init_card_error(card, error);
        return error;
    }

    if (vpibits!=0 && vpibits!=1 && vpibits!=2 && vpibits!=8) {
        printk ("nicstar%d: incorrect number of vpi bits = %d. Should be 0, 1, 2, or 8.\n", card->index, vpibits);
        return 1;
    }
    ns_cfg = cfg[vpibits];

    if ((card = kmalloc(sizeof (ns_dev), GFP_KERNEL)) == NULL) {
        printk("nicstar%d: can't allocate memory for device structure.\n", i);
        error = 2;
        ns_init_card_error(card, error);
        return error;
    }
    /* clear the ns_dev structure */
    memset (card, 0, sizeof (ns_dev));

    cards[i] = card;
    spin_lock_init(&card->int_lock);
    spin_lock_init(&card->res_lock);

    pci_set_drvdata(pcidev, card);

    card->index = i;
    card->atmdev = NULL;
    card->pcidev = pcidev;
    membase = pci_resource_start(pcidev, 1);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
    card->membase = (unsigned long)ioremap(membase, NS_IOREMAP_SIZE);
#else
    card->membase = ioremap(membase, NS_IOREMAP_SIZE);
#endif

    if (card->membase == 0) {
        printk("nicstar%d: can't ioremap() membase.\n", card->index);
        error = 3;
        ns_init_card_error(card, error);
        return error;
    }
    PRINTK("nicstar%d: membase at 0x%lx\n", i, (unsigned long) card->membase);

#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,0)
    pci_set_master(pcidev);
#else
    {
        unsigned short  pci_command;

        if (pci_read_config_word(pcidev, PCI_COMMAND, &pci_command) != 0) {
            printk("nicstar%d: can't read PCI_COMMAND.\n", i);
            error = 4;
            ns_init_card_error(card, error);
            return error;
        }
        pci_command |= (PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);
        if (pci_write_config_word(pcidev, PCI_COMMAND, pci_command) != 0) {
            printk("nicstar%d: can't write PCI_COMMAND.\n", i);
            error = 5;
            ns_init_card_error(card, error);
            return error;
        }
    }
#endif
    if (pci_read_config_byte(pcidev, PCI_LATENCY_TIMER, &pci_latency) != 0) {
        printk("nicstar%d: can't read PCI latency timer.\n", i);
        error = 6;
        ns_init_card_error(card, error);
        return error;
    }
#ifdef NS_PCI_LATENCY
    if (pci_latency < NS_PCI_LATENCY) {
        PRINTK("nicstar%d: setting PCI latency timer to %d.\n", i, NS_PCI_LATENCY);
        for (j = 1; j < 4; j++)  {
            if (pci_write_config_byte(pcidev, PCI_LATENCY_TIMER, NS_PCI_LATENCY) != 0)
                break;
        }
        if (j == 4) {
            printk("nicstar%d: can't set PCI latency timer to %d.\n", i, NS_PCI_LATENCY);
            error = 7;
            ns_init_card_error(card, error);
            return error;
        }
    }
#endif /* NS_PCI_LATENCY */

    /* workaround for SAR bugs #14 and #21*/
    if ((pci_write_config_word(pcidev, 0x40, 0)) || 
        (pci_write_config_byte(pcidev, 0x30, 1))) {
        printk("nicstar%d: can't write to configuration space.\n", i);
        error = 7;
        ns_init_card_error(card, error);
        return error;
    }

    /* Clear timer overflow */
    data = readl(card->membase + STAT);
    if (data & NS_STAT_TMROF)
        writel(NS_STAT_TMROF, card->membase + STAT);

    /* Software reset */
    writel(NS_CFG_SWRST, card->membase + CFG);
    NS_DELAY;
    writel (0x00000000, card->membase + CFG);
    data = readl(card->membase + GP);
    PRINTK("nicstar%d: GP read (0x%08X).\n", i, data);

    /* Card hardware initialization */
    error = ns_phys_init (card, &ns_cfg);  
    if (error) {
        ns_init_card_error(card, error);
        return error;
    }

    writel (ns_cfg, card->membase + CFG);   /* init rct and tct sizes */

    /* Allocate IRQ */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 22)
    if (request_irq(pcidev->irq, &ns_irq_handler, 
                       IRQF_DISABLED|IRQF_SHARED, card->name, card) != 0) {
#else
    if (request_irq(pcidev->irq, &ns_irq_handler, 
                        SA_INTERRUPT | SA_SHIRQ, card->name, card) != 0) {
#endif

    printk("nicstar%d: can't allocate IRQ %d for %s.\n", 
                        card->index, pcidev->irq, card->name);
        error = 9;
        ns_init_card_error(card, error);
        return error;
    }

    /* Set the VPI/VCI MSB mask (must be zero to receive OAM cells) */
    card->vpm = (vcibase>>card->vcibits) |
                 ((vpibase>>card->vpibits) << (16 - card->vcibits));
    card->vpm &= 0x7fff;
    writel(card->vpm, card->membase + VPM);
    printk ("nicstar%d: vpibits=%d, vpibase=%d, vcibase=%d, VPM=0x%x\n", i, 
                                                vpibits, vpibase, vcibase, card->vpm);
    /* Initialize TSQ */
    card->tsq.org = (char*)pci_alloc_consistent(card->pcidev, 
                            NS_TSQSIZE + NS_TSQ_ALIGNMENT, &card->tsq.org_pa);

    if (card->tsq.org == NULL) {
        printk("nicstar%d: can't allocate TSQ.\n", i);
        error = 10;
        ns_init_card_error(card, error);
        return error;
    }
    card->tsq.base_pa = ALIGN_BUS_ADDR(card->tsq.org_pa, NS_TSQ_ALIGNMENT);
    card->tsq.base = (ns_tsi *) (card->tsq.org + (card->tsq.base_pa - card->tsq.org_pa));
    card->tsq.next = card->tsq.base;
    for (j = 0; j < NS_TSQ_NUM_ENTRIES; j++)
             ns_tsi_init(card->tsq.base + j);

    writel(0x00000000, card->membase + TSQH);
    writel(card->tsq.base_pa, card->membase + TSQB);

    /* Initialize RSQ */
    card->rsq.org = (char*)pci_alloc_consistent(card->pcidev, 
                            NS_RSQSIZE + NS_RSQ_ALIGNMENT, &card->rsq.org_pa);
    if (card->rsq.org == NULL) {
        printk("nicstar%d: can't allocate RSQ.\n", i);
        error = 11;
        ns_init_card_error(card, error);
        return error;
    }
    card->rsq.base_pa = ALIGN_BUS_ADDR(card->rsq.org_pa, NS_RSQ_ALIGNMENT);
    card->rsq.base = (ns_rsqe *) (card->rsq.org + (card->rsq.base_pa - card->rsq.org_pa));
    card->rsq.next = card->rsq.base;
    card->rsq.last = card->rsq.base + (NS_RSQ_NUM_ENTRIES - 1);
    for (j = 0; j < NS_RSQ_NUM_ENTRIES; j++)
             ns_rsqe_init(card->rsq.base + j);

    writel(0x00000000, card->membase + RSQH);
    writel(card->rsq.base_pa, card->membase + RSQB);

    /* Initialize TCT */
    u32d[0] = 0x00000000;
    u32d[1] = 0x00000000;
    u32d[2] = 0x00000000;
    u32d[3] = 0x00000000;
    for (j = 0; j < (2 *card->rct_size); j++)
             ns_write_sram(card, j *4, u32d, 4);

    /* Initialize RCT. AAL type is set on opening the VC. */
#ifdef NS_RCQ_SUPPORT
    u32d[0] = NS_RCTE_RAWCELLINTEN;
#else 
    u32d[0] = 0x00000000;
#endif 
    u32d[1] = 0x00000000;
    u32d[2] = 0x00000000;
    u32d[3] = 0xFFFFFFFF;
    for (j = 0; j < card->rct_size; j++)
             ns_write_sram(card, card->rct + (j *4), u32d, 4);

    /* Initialize RAWHND */
    card->raw_hnd = pci_alloc_consistent(card->pcidev, 2 * sizeof(u32),
                                                &card->raw_hnd_pa);
    if (!card->raw_hnd) {
        printk("nicstar%d: can't allocate raw_hnd.\n", i);
        error = 12;
        ns_init_card_error(card, error);
        return error;
    }
    card->raw_hnd[0] = 0;
    writel(card->raw_hnd_pa, card->membase + RAWHND);


    /* Initialize the vc map */
    card->vcmap = (vc_map *)__get_free_pages (GFP_KERNEL, 
                            get_order(card->rct_size *sizeof(vc_map)));
    if (card->vcmap  == NULL) {
        printk("nicstar%d: can't allocate memory for VCs map.\n", i);
        error = 13;
        ns_init_card_error(card, error);
        return error;
    }

    memset(card->vcmap, 0, card->rct_size *sizeof (vc_map));
    for (j = 0; j<card->rct_size; j++)
        card->vcmap[j].conn = j;     /* initialize the connection number */

    for (j = 0; j < NS_SCD_NUM; j++)
             card->scd2vc[j] = NULL;

    /* Initialize ABRVBR TABLES */
    writel(NS_ABRVBRSIZE | (card->dtst<<2), card->membase + ABRSTD);
    for (j = 0; j < NS_DTST_SIZE; j++) {
        ns_write_sram(card, card->dtst + j, u32d, 1);
    }

    /* Initialize the Rate Table */
    abr_vbr_rate_tables = abr_vbr_rate_tables_155Mb;

    for (j = 0; j < abr_vbr_rate_tables_len; j++) {
        ns_write_sram(card, (card->rate + j), &abr_vbr_rate_tables[j], 1);
    }
    writel(card->rate << 2, card->membase + RTBL);

    /* Initialize the TST (use only one) */
    card->tst_free_entries = card->tst_num;
    data = NS_TST_OPCODE_VARIABLE;
    for (j = 0; j < card->tst_num; j++)
             ns_write_sram(card, card->tst + j, &data, 1);

    data = ns_tste_make(NS_TST_OPCODE_END, card->tst<<2);
    ns_write_sram(card, card->tst + card->tst_num, &data, 1);

    writel(card->tst << 2, card->membase + TSTB);

    /* Initialize SCQ0, the UBR SCQ for unspecified PCR */
    card->scq0 = get_scq(card, UBR_SCQ, card->scd_ubr0);
    if (card->scq0 == (scq_info *) NULL) {
        printk("nicstar%d: can't get SCQ0.\n", i);
        error = 14;
        ns_init_card_error(card, error);
        return error;
    }
    u32d[0] = card->scq0->base_pa;
    u32d[1] = (u32) 0x00000000;
    u32d[2] = (u32) 0xffffffff;
    u32d[3] = (u32) 0x00000000;
    ns_write_sram(card, card->scd_ubr0, u32d, 4);
    for (j = 1; j < 3; j++) {
        u32d[0] = (u32)0x00000000;
        u32d[1] = (u32)0x00000000;
        u32d[2] = (u32)0x00000000;
        u32d[3] = (u32)0x00000000;
        ns_write_sram(card, card->scd_ubr0 + 4 *j, u32d, 4);
    }
    card->scq0->scd = card->scd_ubr0;
    PRINTK("nicstar%d: SCQ0 base at 0x%lx.\n", i, (unsigned long)card->scq0->base);

#ifdef NS_OAM_SUPPORT
    /* Initialize oam transmit queue */
    skb_queue_head_init(&card->oam_tq);
#endif

    /* Initialize RxFIFO */
    writel((card->rxfifo << 2) | NS_RXFD_SIZE, card->membase + RXFD);

    /* Initialize free buffer queues registers #34-#45 */
    writel(NS_FBQ0_WP | NS_FBQ0_RP, card->membase + FBQP0);
    writel(NS_FBQ1_WP | NS_FBQ0_RP, card->membase + FBQP1);
    writel(NS_FBQ2_WP | NS_FBQ0_RP, card->membase + FBQP2);
    writel(NS_FBQ3_WP | NS_FBQ0_RP, card->membase + FBQP3);
    writel(NS_B0THLD | NS_B0NITHLD | NS_B0CITHLD | NS_B0SIZE, 
                                                    card->membase + FBQS0);
    writel(NS_B1THLD | NS_B1NITHLD | NS_B1CITHLD | NS_B1SIZE, 
                                                    card->membase + FBQS1);
    writel(NS_B2THLD | NS_B2NITHLD | NS_B2CITHLD, card->membase + FBQS2);
    writel(NS_B3THLD | NS_B3NITHLD | NS_B3CITHLD, card->membase + FBQS3);

    /* Initialize buffer levels */
    card->bnr[0].min = MIN_B0;
    card->bnr[0].init = NUM_B0;
    card->bnr[0].max = MAX_B0;
    card->bnr[1].min = MIN_B1;
    card->bnr[1].init = NUM_B1;
    card->bnr[1].max = MAX_B1;
    card->iovnr.min = MIN_IOVB;
    card->iovnr.init = NUM_IOVB;
    card->iovnr.max = MAX_IOVB;
    card->hbnr.min = MIN_HB;
    card->hbnr.init = NUM_HB;
    card->hbnr.max = MAX_HB;
    card->skb_w[0] =  card->skb_w[1] = NULL;
    card->fbie = 1; /* To prevent push_rxbufs from enabling the interrupt */

    /* Pre-allocate some huge buffers */
    PRINTK("nicstar%d: pre-allocate huge buffers.\n", i);
    skb_queue_head_init(&card->hbpool.queue);
    card->hbpool.count = 0;
    for (j = 0; j < NUM_HB; j++) {
        struct sk_buff  *hb;

        hb = alloc_skb(NS_HBUFSIZE, GFP_KERNEL);
        if (hb == NULL) {
            printk("nicstar%d: can't allocate %dth of %d huge buffers.\n", 
                   i, j, NUM_HB);
            break;
        }
        NS_SKB(hb)->buftype = BUF_HB;
        skb_queue_tail(&card->hbpool.queue, hb);
        card->hbpool.count++;
    }

    /* Allocate large buffers : b1 */
    skb_queue_head_init(&card->bpool[1].queue);
    card->bpool[1].count = 0;                           /* Not used */
    for (j = 0; j < NUM_B1; j++) {
        struct sk_buff  *lb;

        lb = dev_alloc_skb(NS_B1SKBSIZE);
        if (lb == NULL) {
            printk("nicstar%d: can't allocate %dth of %d large buffers.\n", 
                   i, j, NUM_B1);
            break;
        }
        skb_queue_tail(&card->bpool[1].queue, lb);
        skb_reserve(lb, NS_B0BUFSIZE);
        push_rxbufs(card, BUF_B1, lb);
        /* Due to the implementation of push_rxbufs() this is 1, not 0 */
        if (j == 1) {
            card->rcbuf = lb;
            card->raw_ch = (ns_rcqe*)lb->data;
            card->raw_ch_pa = NS_SKB(lb)->paddr; 
        }
    }

    /* Verify that push worked well */
    if ((bcount = ns_fbqc_get(readl(card->membase + FBQP1))) < card->bnr[1].min) {
        printk("nicstar%d: Error: allocate %d large buffers but lfbqc = %d only.\n", 
               i, j, bcount);
        error = 16;
        ns_init_card_error(card, error);
        return error;
    }

    /* Allocate small buffers : b0 */
    skb_queue_head_init(&card->bpool[0].queue);
    card->bpool[0].count = 0;                           /* Not used */
    for (j = 0; j < NUM_B0; j++) {
        struct sk_buff  *sb;

        sb = dev_alloc_skb(NS_B0SKBSIZE);
        if (sb == NULL) {
            printk("nicstar%d: can't allocate %dth of %d small buffers.\n", 
                   i, j, NUM_B0);
            break;
        }
        skb_queue_tail(&card->bpool[0].queue, sb);
        skb_reserve(sb, NS_AAL0_HEADER);
        push_rxbufs(card, BUF_B0, sb);
    }
    /* Verify that push worked well */
    if ((bcount = ns_fbqc_get(readl(card->membase + FBQP0))) < card->bnr[0].min) {
        printk("nicstar%d: Error, Allocate %d small buffers but sfbqc = %d only.\n", 
               i, j, bcount);
        error = 16;
        ns_init_card_error(card, error);
        return error;
    }
    /* Allocate iovec buffers */
    skb_queue_head_init(&card->iovpool.queue);
    card->iovpool.count = 0;
    for (j = 0; j < NUM_IOVB; j++) {
        struct sk_buff  *iovb;

        iovb = alloc_skb(NS_IOVBUFSIZE, GFP_KERNEL);
        if (iovb == NULL) {
            printk("nicstar%d: can't allocate %dth of %d iovec buffers.\n", 
                   i, j, NUM_IOVB);
            error = 16;
            ns_init_card_error(card, error);
            return error;
        }
        NS_SKB(iovb)->buftype = BUF_IOV;
        skb_queue_tail(&card->iovpool.queue, iovb);
        card->iovpool.count++;
    }
    card->intcnt = 0;
    card->fbie = 1;

    /* Register device */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
    card->atmdev = atm_dev_register(card->name, &pcidev->dev, &atm_ops, - 1, NULL);
#else
    card->atmdev = atm_dev_register(card->name, &atm_ops, - 1, NULL);
#endif


    if (card->atmdev == NULL) {
        printk("nicstar%d: can't register device %s\n", i, card->name);
        error = 19;
        ns_init_card_error(card, error);
        return error;
    }

    /* Read EEPROM */
    if (ns_parse_mac (mac[i], card->atmdev->esi)) {
        unsigned char   *esi = card->atmdev->esi;

        ns_eeprom_rd (card, NICSTAR_EPROM_PROSUM_MAC_ADDR_OFFSET, esi, 6);
        if (esi[0] != PROSUM_MAC_0 || esi[1] != PROSUM_MAC_1 || esi[2] != PROSUM_MAC_2)
            ns_eeprom_rd (card, NICSTAR_EPROM_IDT_MAC_ADDR_OFFSET, esi, 6);

        printk("nicstar%d: MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", i, 
               esi[0], esi[1], esi[2], esi[3], esi[4], esi[5]);
    }
    card->atmdev->dev_data = card;
    card->atmdev->ci_range.vpi_bits = 8;
    card->atmdev->ci_range.vci_bits = 16;
    card->atmdev->link_rate = card->max_pcr;
    card->atmdev->phy = NULL;

#ifdef CONFIG_ATM_NICSTAR_USE_SUNI
    if (card->max_pcr == ATM_OC3_PCR) {
        PRINTK ("nicstar%d: Init Suni\n", i);
        suni_init(card->atmdev);

        if (fullsdh) {
            /* modify Suni settings to make it generate SDH frames */
            waitfor_notbusy(card);
            writel(NS_CMD_READ_UTILITY | 0x00000146, card->membase + CMD);
            waitfor_notbusy(card);
            data = readl(card->membase + DR0);
            data |= 0x8;           /*  Set S1 bit of TPOP arbitrary pointer MSB */
            waitfor_notbusy(card);
            writel(data, card->membase + DR0);
            writel(NS_CMD_WRITE_UTILITY | 0x00000146, card->membase + CMD);
            printk ("nicstar%d: Set SDH framing\n", i);
        }
    }
#endif 
    if (card->atmdev->phy && card->atmdev->phy->start)
        card->atmdev->phy->start(card->atmdev);


    writel(NS_DEFAULT_CFG | ns_cfg, card->membase + CFG);

    PRINTK("nicstar%d: card->vpibits= %d.\n", i, card->vpibits);
    PRINTK("nicstar%d: card->vcibits= %d.\n", i, card->vcibits);

#ifdef NS_OAM_SUPPORT
    /*init skb queue of OAM cells waiting for transmission */
    NS_INIT_WORK(&card->oam_wq, ns_push_oamcell, (void *)card);
#endif

    num_cards++;
    return error;
}

static void __devinit ns_init_card_error(ns_dev *card, int error)
{
    PRINTK("nicstar%d: ns_init_card_error (error = %d)\n", 
                                            card->index, error);

    if (error >= 19) {
        writel(0x00000000, card->membase + CFG);
    }
    if (error >= 16) {
        struct sk_buff  *sb;

        while ((sb = skb_dequeue(&card->iovpool.queue)) != NULL)
                       dev_kfree_skb_any(sb);

        while ((sb = skb_dequeue(&card->bpool[0].queue)) != NULL)
                     dev_kfree_skb_any(sb);

        free_scq(card, card->scq0, NULL);

        while ((sb = skb_dequeue(&card->bpool[1].queue)) != NULL)
                     dev_kfree_skb_any(sb);

        while ((sb = skb_dequeue(&card->hbpool.queue)) != NULL)
                     dev_kfree_skb_any(sb);
    }
    if (error >= 14) {
        free_pages((unsigned long)card->vcmap, get_order(card->rct_size *sizeof(vc_map)));
    }
    if (error >= 13) {
        pci_free_consistent(card->pcidev, 2 * sizeof(u32), 
                                            card->raw_hnd, card->raw_hnd_pa);
    }

    if (error >= 12) {
        pci_free_consistent(card->pcidev, NS_RSQSIZE + NS_RSQ_ALIGNMENT, 
                                            card->rsq.org, card->rsq.org_pa);
    }
    if (error >= 11) {
        pci_free_consistent(card->pcidev, NS_TSQSIZE + NS_TSQ_ALIGNMENT, 
                                            card->tsq.org, card->tsq.org_pa);
    }
    if (error >= 10) {
        free_irq(card->pcidev->irq, card);
    }
    if (error >= 4) {
        iounmap((void *) card->membase);
    }
    if (error >= 3) {
        pci_disable_device(card->pcidev);
        kfree(card);
    }
}

static scq_info *get_scq(ns_dev *card, int type, u32 scd)
{
    scq_info        *scq;

    scq = (scq_info *) kmalloc(sizeof (scq_info), GFP_KERNEL);
    if (scq == (scq_info *) NULL)
        return (scq_info *) NULL;
                           /* reserve twice the size for alignment purpose*/
    scq->org = pci_alloc_consistent(card->pcidev, 2 * SCQ_SIZE, &scq->org_pa);
        if (scq->org == NULL) {
        kfree(scq);
        return (scq_info *) NULL;
    }
                            /* reserve a table of skbuf pointers */
    scq->skb = (struct sk_buff * *) kmalloc(sizeof (struct sk_buff *) * SCQ_NUM_ENTRIES
                                            , GFP_KERNEL);
    if (scq->skb == NULL) {
        pci_free_consistent(card->pcidev, 2 * SCQ_SIZE, scq->org, scq->org_pa);
        kfree(scq);
        return (scq_info *) NULL;
    }
    memset (scq->skb, 0, sizeof (struct sk_buff *) * SCQ_NUM_ENTRIES);
    scq->scq_type = type;
    scq->base_pa = ALIGN_BUS_ADDR(scq->org_pa, SCQ_SIZE);
    scq->base = (ns_scqe *) (scq->org + (scq->base_pa - scq->org_pa));
    scq->next = scq->base;
    scq->last = scq->base + SCQ_NUM_ENTRIES - 1;
    scq->tail = scq->last;
    scq->scd = scd;
    scq->tbd_count = 0;
    skb_queue_head_init(&scq->waiting);
    init_waitqueue_head(&scq->scqfull_waitq);
    scq->full = 0;
    spin_lock_init(&scq->lock);
    return scq;
}

/* For variable rate SCQ vcc must be NULL */
static void free_scq(ns_dev *card, scq_info *scq, struct atm_vcc *vcc)
{
    int             i;
    struct sk_buff  *skb;

        if (scq==NULL) {
                printk ("nicstar%d: try to free a null SCQ\n", card->index);
                return;
        }

    while ((skb = skb_dequeue(&scq->waiting)))
                dev_kfree_skb_any (skb);

    if (scq == card->scq0) {
        for (i = 0; i < SCQ_NUM_ENTRIES; i++)
        {
            if (scq->skb[i] != NULL) {
                vcc = ATM_SKB(scq->skb[i])->vcc;
                if (vcc->pop != NULL)
                    vcc->pop(vcc, scq->skb[i]);
                else
                    dev_kfree_skb_any(scq->skb[i]);

            }
        }
    } else {
        /* vcc must be != NULL */
        if (vcc == NULL) {
            printk("nicstar%d: free_scq() called with vcc == NULL.", card->index);
            for (i = 0; i < SCQ_NUM_ENTRIES; i++)
                        dev_kfree_skb_any(scq->skb[i]);
        } else {
            for (i = 0; i < SCQ_NUM_ENTRIES; i++)
            {
                if (scq->skb[i] != NULL) {
                    if (vcc->pop != NULL)
                        vcc->pop(vcc, scq->skb[i]);
                    else
                        dev_kfree_skb_any(scq->skb[i]);

                }
            }
        }
    }

    kfree(scq->skb);
    pci_free_consistent(card->pcidev, 2 * SCQ_SIZE, scq->org, scq->org_pa);
    kfree(scq);
}

static int get_index (struct fbq_skb_list *lp)
{
    int i, idx;

    for (idx=lp->idx, i=0; i<NS_SKBL_SIZE; idx=(idx+1)%NS_SKBL_SIZE, i++) {
        if (lp->skb[idx] == NULL) {
            lp->idx = (idx+1)%NS_SKBL_SIZE;
            return idx;
        }
    }
    return -1;     /* no handle available */
}

/*
 * Add one buffer into the SAR Free Buffer Queue 
 * The sk_buff passed must contain the small or large buffer cast to u32.
 * These buffers are filled with incoming data and dequeued automatically
 * by the SAR. 
 */
static void push_rxbufs(ns_dev *card, char type, struct sk_buff *skb)
{

    int index1, index2;
    u32 paddr1 = 0, paddr2 = 0;
    u32 fbqp[2];
    int tn = (type == BUF_B0)? 0: 1;
    unsigned long flags;
    struct sk_buff *skb2 = card->skb_w[tn];
    struct fbq_skb_list *lp = card->fsl+tn;

    PRINTK("nicstar%d: push_rxbufs ()\n", card->index);

    NS_SKB(skb)->buftype = type;        /* mark pool type into cb */

    fbqp[0] = readl(card->membase + FBQP0);
    fbqp[1] = readl(card->membase + FBQP1);
    card->bfqc[0] = ns_fbqc_get(fbqp[0]);
    card->bfqc[1] = ns_fbqc_get(fbqp[1]);

    if (skb2 == NULL) {
        card->skb_w[tn] = skb;   /* only one buffer, store it temporarily */
    }
    else {
        if (card->bfqc[tn] >= card->bnr[tn].max) {
            /* The queue is full, we free the 2 skb's and give up */
            PRINTK("nicstar%d: freing %s buffers.\n", card->index, tn==0? "small": "large");
            goto free_skbs;
        }

        /* look for two buffer handles */
        index1 = get_index(lp);
        index2 = get_index(lp);
        if ((index1 < 0) || (index2 < 0)) {
            printk ("nicstar%d: cannot get handles into push_rxbufs\n", card->index);
            goto free_skbs;
        }

        /* OK we can push the two buffers */
        card->skb_w[tn] = NULL;

        paddr1 = pci_map_single(card->pcidev, skb->data, skb_end_ptr(skb) - skb->data, PCI_DMA_FROMDEVICE);
        paddr2 = pci_map_single(card->pcidev, skb2->data, skb_end_ptr(skb2) - skb2->data, PCI_DMA_FROMDEVICE);

        if (paddr1==0 || paddr2==0) {
            PRINTK("nicstar%d: paddr1 = 0x%x, paddr2 = 0x%x.\n", card->index, paddr1, paddr2);
            goto free_skbs;
        }

        card->bfqc[tn] += 2;

        /* record sk_buff addresses into card structure */
        lp->skb[index1] = skb;
        lp->skb[index2] = skb2;

        /* record physical address of data into private part of sk_buff */
        NS_SKB(skb)->paddr = paddr1;
        NS_SKB(skb2)->paddr = paddr2;

        /* Send new buffer information to SAR */
        ns_grab_res_lock(card, flags);
        PRINTK("nicstar%d: Pushing %s buffers at %p/0x%x and %p/0x%x.\n", card->index, 
               tn==0? "small" : "large", skb, paddr1, skb2, paddr2);

        waitfor_notbusy(card);
        writel(paddr2, card->membase + DR3);
        writel((u32)(index2 | (tn<<16)), card->membase + DR2);      /* handle */
        writel(paddr1, card->membase + DR1);
        writel((u32)(index1 | (tn<<16)), card->membase + DR0);      /* handle */
        writel(NS_CMD_WRITE_FREEBUFQ | tn,  card->membase + CMD);
        spin_unlock_irqrestore(&card->res_lock, flags);
    }

    if (!card->fbie && card->bfqc[0] >= card->bnr[0].min && 
                                    card->bfqc[1] >= card->bnr[1].min) {
        card->fbie = 1;
        writel((readl(card->membase + CFG) | NS_CFG_FBIE), card->membase + CFG);
    }
    return;

free_skbs:
    if (skb) {
        if (paddr1)
            pci_unmap_single(card->pcidev, paddr1, skb_end_ptr(skb) - skb->data, PCI_DMA_FROMDEVICE);

        SKB_UNLINK(skb, &card->bpool[tn].queue);
        dev_kfree_skb_any(skb);
    }
    if (skb2) {
        if (paddr2)
            pci_unmap_single(card->pcidev, paddr2, skb_end_ptr(skb2) - skb2->data, PCI_DMA_FROMDEVICE);    

        SKB_UNLINK(card->skb_w[tn], &card->bpool[tn].queue);
        dev_kfree_skb_any(card->skb_w[tn]);
        card->skb_w[tn] = NULL;
    }
    return;

}
#define INT_FLAGS ( NS_STAT_TSIF  | \
                    NS_STAT_TXICP | \
                    NS_STAT_TSQF  | \
                    NS_STAT_TMROF | \
                    NS_STAT_PHYI  | \
                    NS_STAT_RSQF  | \
                    NS_STAT_EOPDU | \
                    NS_STAT_RAWCF | \
                    NS_STAT_FBQ0A | \
                    NS_STAT_FBQ1A | \
                    NS_STAT_RSQAF)

#define  CLEAR_FLAGS ( NS_STAT_TSIF | NS_STAT_TXICP | NS_STAT_TMROF \
                        | NS_STAT_PHYI | NS_STAT_EOPDU | NS_STAT_RAWCF) 

#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,19)
static irqreturn_t ns_irq_handler(int irq, void *dev_id)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,0)  
static irqreturn_t ns_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
#else
static void ns_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    u32             stat_r, cfg_r;
    ns_dev          *card;
    struct atm_dev  *dev;
    unsigned long   flags;
    int iloop = 0;

    card = (ns_dev *) dev_id;
    dev = card->atmdev;
    ns_grab_int_lock(card, flags);
    stat_r = readl(card->membase + STAT);

    if (!(stat_r & INT_FLAGS)) {
        spin_unlock_irqrestore(&card->int_lock, flags);
        return IRQ_NONE;          /* not for us */
    }

    cfg_r = readl(card->membase + CFG);

    PRINTK("nicstar%d: ns_irq_handler ()\n", card->index);
    XPRINTK("    cfg= 0x%08x, stat= 0x%08x\n", cfg_r, stat_r);

    card->intcnt++;

    while (stat_r & INT_FLAGS) {
        if ( iloop++ > 10)
            break;                              /* break infinite loop */

        writel(stat_r & CLEAR_FLAGS, card->membase + STAT);   /* clear the interrupt */

        /* Transmit Status Indicator has been written to T. S. Queue */
        if (stat_r & NS_STAT_TSIF) {
            PRINTK("nicstar%d: TSI interrupt\n", card->index);
            process_tsq(card);
        }
        /* Transmit Status Queue 7/8 full */
        if (stat_r & NS_STAT_TSQF) {
            PRINTK("nicstar%d: TSQ full.\n", card->index);
            process_tsq(card);
        }
        /* PHY device interrupt signal active */
        if (stat_r & NS_STAT_PHYI) {
            PRINTK("nicstar%d: PHY interrupt.\n", card->index);
            if (dev->phy && dev->phy->interrupt) {
                dev->phy->interrupt(dev);
            }
        }
        /* Receive Status Queue is full */
        if (stat_r & NS_STAT_RSQF) {
            printk("nicstar%d: RSQ full.\n", card->index);
            process_rsq(card);
        }
        /* Complete CS-PDU received */
        if (stat_r & NS_STAT_EOPDU) {
            PRINTK("nicstar%d: End of CS-PDU received.\n", card->index);
            process_rsq(card);
        }
        /* Raw cell received */
        if (stat_r & NS_STAT_RAWCF) {
#ifdef NS_RCQ_SUPPORT
            PRINTK("nicstar%d: Raw cell received\n", card->index);
            ns_rawc (card);
#else
            printk ("nicstar%d: Unexpected raw cell received interrupt\n", card->index);
#endif
        }

        /* Small buffer queue attention */
        /* Happens only when too much traffic. try to allocate more buffers */
        if (stat_r & NS_STAT_FBQ0A) {
            int             i;
            struct sk_buff  *sb;
            u32             fbqp0;

            PRINTK("nicstar%d: free buffer queue 0 attention.\n", 
                    card->index);
            for (i = 0; i < card->bnr[0].min; i++)
            {
                sb = dev_alloc_skb (NS_B0SKBSIZE);
                if (sb == NULL) {
                    /* if no more buffers available. Disable Q0A interrupt and give up */
                    /* this interrupt will be enabled again when buffers are recycled */
                    writel(readl(card->membase + CFG)& ~NS_CFG_FBIE, card->membase + CFG);
                    card->fbie = 0;
                    break;
                }
                skb_queue_tail(&card->bpool[0].queue, sb);
                skb_reserve(sb, NS_AAL0_HEADER);
                push_rxbufs(card, BUF_B0, sb);
            }
            /*card->bfqc[0] = i;*/
            fbqp0 = readl(card->membase + FBQP0);
            card->bfqc[0] = ns_fbqc_get(fbqp0);
            process_rsq(card);
        }

        /* Large buffer queue attention */
        /* Happens only when too much traffic. try to allocate more buffers */
        if (stat_r & NS_STAT_FBQ1A) {
            int             i;
            struct sk_buff  *lb;
            u32             fbqp1;

            PRINTK("nicstar%d: free buffer queue 1 attention.\n", 
                    card->index);
            for (i = 0; i < card->bnr[1].min; i++)
            {
                lb = dev_alloc_skb(NS_B1SKBSIZE);
                if (lb == NULL) {
                    /* if no more buffers available. Disable Q0A interrupt and give up */
                    /* this interrupt will be enabled again when buffers are recycled */
                    writel(readl(card->membase + CFG)& ~NS_CFG_FBIE, card->membase + CFG);
                    card->fbie = 0;
                    break;
                }
                skb_queue_tail(&card->bpool[1].queue, lb);
                skb_reserve(lb, NS_B0BUFSIZE);
                push_rxbufs(card, BUF_B1, lb);
            }
            fbqp1 = readl(card->membase + FBQP1);
            card->bfqc[1] = ns_fbqc_get(fbqp1);
            process_rsq(card);
        }

        /* Receive Status Queue is 7/8 full */
        if (stat_r & NS_STAT_RSQAF) {
            PRINTK("nicstar%d: RSQ 7/8 full.\n", card->index);
            process_rsq(card);
        }

        /* Incomplete CS-PDU has been transmitted */
        if (stat_r & NS_STAT_TXICP) {
            printk("nicstar%d: Incomplete CS-PDU transmitted.\n", card->index);
        }
        /* Timer overflow */
        if (stat_r & NS_STAT_TMROF) {
            PRINTK("nicstar%d: Timer overflow.\n", card->index);
        }
 
        stat_r = readl(card->membase + STAT);
    }

    XPRINTK("nicstar%d: end of interrupt service\n", card->index);

    spin_unlock_irqrestore(&card->int_lock, flags);

#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,0)
    return IRQ_HANDLED;
#endif
}

#ifdef NS_OAM_SUPPORT 
static int ns_oam_type (u32 pti, u32 vci)
{
    if ( (pti == ATM_PTI_SEGF5) &&
         (vci != 0x00) && (vci != 0x03) &&
         (vci != 0x04) && (vci != 0x06) && (vci != 0x07) ) {
        PRINTK ("nicstar: OAM F5 segment");
        return OAM_F5_SEG;
    }
    else if ( (pti == ATM_PTI_E2EF5) &&
         (vci != 0x00) && (vci != 0x03) &&
         (vci != 0x04) && (vci != 0x06) && (vci != 0x07) ) {
        PRINTK ("nicstar: OAM F5 end-to-end");
        return OAM_F5_E2E;
    }
    else if (vci == 0x03) {
        PRINTK ("nicstar: OAM F4 segment");
        return OAM_F4_SEG;
    }
    else if (vci == 0x04) {
        PRINTK ("nicstar: OAM F4 end-to-end");
        return OAM_F4_E2E;
    }
    return 0;
}
#endif

#ifdef NS_PRINT_RAW_CELL_MESSAGE
#define rcmes()  printk("nicstar%d: raw cell received on VC %d.%d, header = 0x%08x\n",  \
                            card->index, vpi, vci, rawcell->word_1)
#else
#define rcmes()
#endif

/* 
 * Process raw cells 
 */
#ifdef NS_RCQ_SUPPORT
static int ns_process_rcqe (ns_dev *card, ns_rcqe *rawcell)
{
    u32 vci, vpi, pti, conn;
    vc_map *vc;
    struct atm_vcc *vcc;
    struct sk_buff  *sb;

    vpi = (le32_to_cpu(rawcell->word_1) & 0x0FF00000) >> 20;
    vci = (le32_to_cpu(rawcell->word_1) & 0x000FFFF0) >> 4;
    pti = (le32_to_cpu(rawcell->word_1) & 0x0E) >> 1;

#ifdef NS_OAM_SUPPORT
    if (ns_oam_type(pti, vci)) {
        u32 conn = ns_find_conn (card, vpi, vci);
        vc_map *vc = &(card->vcmap[conn]);

        PRINTK(" cell received on %d.%d\n", vpi, vci);

        if (!vc->rx) {
            rcmes();      /* we don't support oam if connection is not open */
            return 0;
        }

        rawcell->word_2 = conn;
        if (ns_process_oam_cell (card, rawcell))
            rcmes();
        return 0;
    }
#endif

#ifdef NS_AAL0_TO_RCQ
    conn = ns_find_conn (card, vpi, vci);
    vc = &(card->vcmap[conn]);
    if (!vc->rx) {
        rcmes();      /* connection is not open */
        return 0;
    }

    vcc = vc->rx_vcc;

    if (vcc->qos.aal != ATM_AAL0) {
        rcmes();       /* give up if not AAL0 */
        return 0;
    }

    sb = dev_alloc_skb(NS_B0SKBSIZE);
    if (!sb) {
        printk("nicstar%d: Can't allocate buffers for aal0 raw cell.\n", card->index);
        atomic_add(1, &vcc->stats->rx_drop);
        return -1;
    }
    if (!atm_charge(vcc, sb->truesize)) {
        PRINTK("nicstar%d: atm_charge() dropped aal0 raw cell.\n", 
                    card->index);
        dev_kfree_skb_any(sb);
        return -1;
    }

    *((u32 *)sb->data) = le32_to_cpu(rawcell->word_1);
    skb_put(sb, NS_AAL0_HEADER);
    memcpy(skb_put(sb, ATM_CELL_PAYLOAD), rawcell->payload, ATM_CELL_PAYLOAD);
    ATM_SKB(sb)->vcc = vcc;
    TIMESTP(sb);
    vcc->push(vcc, sb);
    atomic_inc(&vcc->stats->rx);
    return 0;
#endif
    rcmes();
    return 0;
}

/* Raw cell receipt */
static void ns_rawc(ns_dev *card)
{
    ns_rcqe      *rawcell;
    u32          tail;
    u32         handle, index, listNum;

    if (card->raw_hnd[0] == 0 || card->rcbuf == NULL)
        return;

    tail = le32_to_cpu(card->raw_hnd[0]) + NS_RCQE_SIZE;

//    while (pci_map_single(card->pcidev, card->raw_ch, 
//                                        NS_RCQE_SIZE, PCI_DMA_FROMDEVICE) != tail) {
    while (card->raw_ch_pa != tail) {
        rawcell = card->raw_ch;

        if (ns_rcqe_islast(rawcell)) {
            struct sk_buff  *oldbuf = card->rcbuf;
            handle = ns_rcqe_nextbufhandle(rawcell);   /* get next buffer type and index */
            index = handle & (NS_SKBL_SIZE-1);
            listNum = (handle >> 16) & 1;

            PRINTK("nicstar: listNum= %d, handle= 0x%08x, index= 0x%08x\n", listNum, handle, index);

            card->rcbuf = card->fsl[listNum].skb[index];      /* get sk_buff address */
            card->fsl[listNum].skb[index] = NULL;

            pci_unmap_single(card->pcidev, NS_SKB(oldbuf)->paddr, 
                            skb_end_ptr(oldbuf) - oldbuf->data, PCI_DMA_FROMDEVICE);

            recycle_rx_buf(card, oldbuf);

            if (card->rcbuf == NULL) {
                printk ("nicstar%d: raw cell queue error. Stop receiving raw cells.\n", card->index);
                return;
            }

            card->raw_ch = (ns_rcqe*)(card->rcbuf->data);
            card->raw_ch_pa = NS_SKB(card->rcbuf)->paddr;
        }
        else {
            ns_process_rcqe (card, rawcell);
            card->raw_ch++;
            card->raw_ch_pa += NS_RCQE_SIZE;
        }
    }
}
#endif /* NS_RCQ_SUPPORT */

/* Compute (scr<<16)/pcr */
static u16 scr_on_pcr(u32 scr, u32 pcr)
{
    u32 q = 0;

    scr <<= 8;
    q = scr/pcr;
    scr = (scr % pcr) << 8;
    return  (q << 8) | (scr/pcr) ;
}

/* translate rate to ATM Forum rate 
 * 2^e (1+m/512) = rate   => m = rate * 2^(9-e) - 512
 */
static u16 u32_to_afr(u32 rate)
{
    u16 m, e;
    u32 mask;

    if (rate == 0) return 0;
    for (mask = 0x80000000, e = 31; mask; mask >>= 1, e--)
            if (rate & mask) break;
    m = e >= 9 ? rate >> (e-9): rate << (9-e);
    m -= 512;

    return (1<<14 | e<<9 | m);
}
/* 
 * First transform rate to ATM fixed point format (afr)
 * Then find the index corresponding to afr into log_to_conv table
 */
static u8 ns_rate2log (ns_dev *card, u32 rate)
{
    u16 afr_rate = u32_to_afr(rate), afr_value;
    u32 data;
    u8 lower =0, upper =255, mid;
    u32 * rate_table;

    rate_table = abr_vbr_rate_tables_155Mb;
 
    /* 
     * Could use the rate2log table but this content 
     * research gives better precision 
     */
    while (1) {
        mid = (u8) ((lower + upper) >> 1);
        data = rate_table [mid];
        afr_value = (u16) (data >> 17);  /* extract cps from table */

        if ((afr_rate == afr_value) || (upper <= lower))
            break;
        else if (afr_rate > afr_value)
            lower = mid + 1;
        else
            upper = mid - 1;
    }
    if (afr_value > afr_rate)
        mid--;

    return (mid);
}

static int ns_init_tx_cbr (ns_dev *card, vc_map *vc, struct atm_vcc *vcc, int conn)
{
    scq_info        *scq;
    int             j;
    u32             u32d[8];
    int             tcr, tcra;      /* target cell rate, and absolute value */
    int             n = 0;          /* Number of entries in the TST */
    u32             tmpl, modl;
    int             scdi = 0;     /* SCD Index */

    PRINTK("nicstar%d: ns_init_tx_cbr ()\n", card->index);
    XPRINTK ("    pcr= 0x%x, max_pcr= 0x%x, min_pcr= 0x%x\n", 
                        vcc->qos.txtp.pcr, vcc->qos.txtp.max_pcr, vcc->qos.txtp.min_pcr);

    /* Check requested cell rate and availability of SCD */
    if (vcc->qos.txtp.max_pcr == 0 && vcc->qos.txtp.pcr == 0 && 
        vcc->qos.txtp.min_pcr == 0) {
        printk("nicstar%d: trying to open a CBR vc with cell rate = 0 \n", 
                card->index);
        return - EINVAL;
    }
    /* Cell rate computation */
    tcr = atm_pcr_goal(& (vcc->qos.txtp));
    tcra = tcr >= 0 ? tcr : - tcr;
    PRINTK("nicstar%d: target cell rate = %d.\n", card->index, tcr); 
    tmpl = (u32)tcra * (u32)card->tst_num;
    modl = tmpl % card->max_pcr;
    n = (int)(tmpl / card->max_pcr);
    if (tcr > 0) {
        if (modl > 0)
            n++;
    }
    else if (tcr == 0) {
        if ((n = (card->tst_free_entries - NS_TST_RESERVED)) <= 0) {
            printk("nicstar%d: no CBR bandwidth free.\n", card->index);
            return - EINVAL;
        }
    }
    if (n == 0) {
        printk ("nicstar%d: selected bandwidth < granularity.\n", card->index);
        return - EINVAL;
    }
    if (n > (card->tst_free_entries - NS_TST_RESERVED)) {
        printk ("nicstar%d: not enough free CBR bandwidth.\n", card->index);
        return - EINVAL;
    }
    else
        card->tst_free_entries -= n;

    PRINTK("nicstar%d: writing %d tst entries.\n", card->index, n);

    /* get SCD */
    for (scdi = 0; scdi < NS_SCD_NUM; scdi++) {
        if (card->scd2vc[scdi] == NULL) {
            card->scd2vc[scdi] = vc;
            break;
        }
    }
    if (scdi == NS_SCD_NUM) {
        printk("nicstar%d: no SCD available for CBR channel.\n", card->index);
        card->tst_free_entries += n;
        return - EBUSY;
    }
    vc->adr_scd = card->scd + scdi * NS_SCD_ENTRY_SIZE;

    /* get SCQ */
    scq = get_scq(card, CBR_SCQ, vc->adr_scd);
    if (scq == (scq_info *) NULL) {
        printk("nicstar%d: can't get Fixed Rate SCQ.\n", card->index);
        card->scd2vc[scdi] = NULL;
        card->tst_free_entries += n;
        return - ENOMEM;
    }
    vc->scq = scq;

    /* set up SCD */
    u32d[0] = scq->base_pa;
    u32d[1] = (u32)0x00000000;
    u32d[2] = (u32)0xffffffff;
    u32d[3] = (u32)0x00000000;
    ns_write_sram(card, vc->adr_scd, u32d, 4);
    for (j = 1; j < 3; j++)                   /* initialize the cache */
    {
        u32d[0] = (u32)0x00000000;
        u32d[1] = (u32)0x00000000;
        u32d[2] = (u32)0x00000000;
        u32d[3] = (u32)0x00000000;
        ns_write_sram(card, vc->adr_scd + 4 *j, u32d, 4);
    }
    PRINTK("nicstar%d: ns_open (), CBR scd written at:0x%x\n", card->index, vc->adr_scd);

    /* set up TCT */
    u32d[0] = NS_TCTE_TYPE_CBR | vc->adr_scd;
    u32d[1] = 0x00000000;
    u32d[2] = 0x00000000;
    u32d[3] = 0x00000000;
    ns_write_sram(card, (conn * 8), u32d, 4);
    u32d[4] = 0x00000000;
    u32d[5] = 0x00000000;
    u32d[6] = 0x00000000;
    u32d[7] = 0x00000000;
    ns_write_sram(card, (conn * 8) + 4, u32d+4, 4);

    PRINTK("nicstar: a (type CBR)TCT entry written: word_1= 0x%x.\n", u32d[0]);

    /* set up TST */
    fill_tst(card, n, vc, conn);

    return 0;
}

static int ns_init_tx_ubr (ns_dev *card, vc_map *vc, struct atm_vcc *vcc, int conn)
{
    scq_info        *scq;
    int             j;
    u32             u32d[8];
    u32             max_pcr = 0;
    int             scdi = 0;     /* SCD Index */
    int             fubr0 = 0;

    PRINTK("nicstar%d: ns_init_tx_ubr ()\n", card->index);
    XPRINTK ("    pcr= 0x%x, max_pcr= 0x%x\n", vcc->qos.txtp.pcr, vcc->qos.txtp.max_pcr);

    if (vcc->qos.txtp.pcr == 0 ) {
        PRINTK("nicstar%d: trying to open a UBR vc (unspecified PCR)\n", card->index);
        vc->init_er = 0xff;
        vc->lacr = 0xff;
        fubr0 = 1;
    } 
    else {
        max_pcr = vcc->qos.txtp.max_pcr;
        if ( max_pcr <  vcc->qos.txtp.pcr)
            max_pcr =  vcc->qos.txtp.pcr;

            vc->init_er     = ns_rate2log (card, max_pcr);
            vc->lacr        = ns_rate2log (card, vcc->qos.txtp.pcr);

            XPRINTK ("    max_pcr= 0x%x, init_er= 0x%x, lacr= 0x%x\n", 
                                    max_pcr, vc->init_er, vc->lacr);
            if (vc->init_er == 0xff && vc->lacr == 0xff)
                    fubr0 = 1; 
    }
    if (fubr0) {
        vc->adr_scd = card->scd_ubr0;
        vc->scq = card->scq0;    /* no need to update */
    }
    else {
        /* UBR with specific pcr value. We use a specific SCD */

        /* get SCD */
        for (scdi = 0; scdi < NS_SCD_NUM; scdi++) {
            if (card->scd2vc[scdi] == NULL) {
                card->scd2vc[scdi] = vc;
                break;
            }
        }
        if (scdi == NS_SCD_NUM) {
            printk ("nicstar%d: no SCD available for UBR channel.\n", card->index);
            return - EBUSY;
        }
        vc->adr_scd = card->scd + scdi * NS_SCD_ENTRY_SIZE;
        scq = get_scq (card, UBR_SCQ, vc->adr_scd);
        if (scq == (scq_info *) NULL) {
            printk ("nicstar%d: can't get UBR SCQ.\n", card->index);
            card->scd2vc[scdi] = NULL;
            return - ENOMEM;
        }
        vc->scq = scq;

        /* set up SCD */
        u32d[0] = scq->base_pa;
        u32d[1] = (u32)0x00000000;
        u32d[2] = (u32)0xffffffff;
        u32d[3] = (u32)0x00000000;
        ns_write_sram(card, vc->adr_scd, u32d, 4);
        for (j = 1; j < 3; j++) {           /* initialize the cache */
            u32d[0] = (u32)0x00000000;
            u32d[1] = (u32)0x00000000;
            u32d[2] = (u32)0x00000000;
            u32d[3] = (u32)0x00000000;
            ns_write_sram(card, vc->adr_scd + 4 *j, u32d, 4);
        }
        XPRINTK("nicstar: UBR scd written at:0x%x\n", (int)vc->adr_scd);
    }

    /* finaly create the UBR entry in the TCT */
    u32d[0] = NS_TCTE_TYPE_UBR | vc->adr_scd;
    u32d[1] = 0x00000000;
    u32d[2] = 0x00000000;
    u32d[3] = NS_TCTE_HALT | NS_TCTE_IDLE;
    ns_write_sram(card, (conn *8), u32d, 4);
    u32d[4] = 0x00000000;
    u32d[5] = 0x00000000;
    u32d[6] = 0x00000000;
    u32d[7] = 0x80000000;
    ns_write_sram(card, (conn *8) + 4, u32d+4, 4);

    XPRINTK ("nicstar: a (type UBR) TCT entry written at 0x%lx:\n",
                               (unsigned long)(card->membase + (conn * 8)));
    return 0;
}

/*
 * VBR is normally not supported by Linux. A VBR QoS is defined by 3 parameters
 * PCR, SCR and MBS.  To set a VBR connection, pass the VBR parameters into the
 * standard traffic parameters of the atm_trafprm stucture as following:
 *
 *      traffic_class = ATM_VBR (3),
 *      max_pcr = the needed PCR value, 
 *      pcr =  the needed SCR value,
 *      min_pcr = the needed MBS value (1 to 255)
 */ 

static int ns_init_tx_vbr (ns_dev *card, vc_map *vc, struct atm_vcc *vcc, int conn)
{
    scq_info        *scq;
    int             j;
    u32             u32d[8];
    u16             pcr_token;
    int             scdi = 0;     /* SCD Index */
    int             scr, pcr, mbs;

    PRINTK("nicstar%d: ns_init_tx_vbr ()\n", card->index);
    XPRINTK ("    max_pcr= 0x%x, pcr= 0x%x, min_pcr= 0x%x\n", 
                  vcc->qos.txtp.max_pcr, vcc->qos.txtp.pcr, vcc->qos.txtp.min_pcr);

    mbs = vcc->qos.txtp.min_pcr;   /* use min_pcr as MBS */
    scr = vcc->qos.txtp.pcr;            /* use pcr as SCR */
    pcr = vcc->qos.txtp.max_pcr;     /* use max_pcr as PCR */

    if (mbs <= 0 || mbs > 255) {
        printk("nicstar%d: trying to open a VBR vc with invalid MBS: %d\n", 
                card->index, mbs);
        return -EINVAL;
    } 
    if (scr <= 0 || scr > pcr) {
        printk("nicstar%d: trying to open a VBR vc with SCR null or higher than PCR: %d\n", 
                card->index, scr);
        return -EINVAL;
    }
    if (pcr <= 0 || pcr > card->max_pcr) {
        printk("nicstar%d: trying to open a VBR vc with invalid PCR, %d\n", 
                card->index, pcr);
        return -EINVAL;
    }

    /* get SCD */
    for (scdi = 0; scdi < NS_SCD_NUM; scdi++) {
        if (card->scd2vc[scdi] == NULL) {
            card->scd2vc[scdi] = vc;
            break;
        }
    }
    if (scdi == NS_SCD_NUM) {
        printk ("nicstar%d: no SCD available for VBR channel.\n", card->index);
        return -EBUSY;
    }
    vc->adr_scd = card->scd + scdi * NS_SCD_ENTRY_SIZE;
    scq = get_scq(card, VBR_SCQ, vc->adr_scd);
    if (scq == (scq_info *) NULL) {
        printk ("nicstar%d: can't get VBR SCQ.\n", card->index);
        card->scd2vc[scdi] = NULL;
        return -ENOMEM;
    }
    vc->scq = scq;

    /* set up SCQ */
    u32d[0] = scq->base_pa;
    u32d[1] =  u32d[0] & 0x00001FF0;
    if ((u32d[0]& 0x00001FFF) != u32d[1])
        printk("nicstar%d: error in scq->base address \n", card->index );

    u32d[2] = (u32)0xffffffff;
    u32d[3] = (u32)0x00000000;
    ns_write_sram(card, vc->adr_scd, u32d, 4);
    for (j = 1; j < 3; j++) {           /* initialize the cache */
        u32d[0] = (u32)0x00000000;
        u32d[1] = (u32)0x00000000;
        u32d[2] = (u32)0x00000000;
        u32d[3] = (u32)0x00000000;
        ns_write_sram(card, vc->adr_scd + 4 *j, u32d, 4);
    }
    XPRINTK("nicstar: VBR scd written at:0x%x\n", vc->adr_scd);

    /* create the VBR entry in the TCT */
    pcr_token = scr_on_pcr (scr, pcr);   
    if (pcr_token == 0) 
        pcr_token++;
    vc->init_er = ns_rate2log (card, pcr);
    vc->lacr    = (u32)ns_rate2log (card, scr);

    u32d[0] = NS_TCTE_TYPE_VBR | vc->adr_scd;
    u32d[1] = 0x00000000;
    u32d[2] = NS_TCTE_TSIF;
    u32d[3] = NS_TCTE_HALT | NS_TCTE_IDLE;
    ns_write_sram(card, (conn *8), u32d, 4);

    u32d[4] = 0x7F<<24;                 /* max IdleCount = 0x7f*/  
    u32d[5] = 0x01<<24;                 /* token bucket integer portion=1*/ 
    u32d[6] = (mbs << 16) | pcr_token;             /* icr is mbs */
    u32d[7] = 0x00000000;
    ns_write_sram(card, (conn *8) + 4, u32d+4, 4);

    XPRINTK ("nicstar: a (type VBR)TCT entry written at 0x%lx:\n",
                                    (unsigned long)(card->membase + (conn * 8)));

    return 0;
}

static int ns_init_tx_abr (ns_dev *card, vc_map *vc, struct atm_vcc *vcc, int conn)
{
    scq_info        *scq;
    int             j;
    u32             u32d[8];
    u32             air_table, rdf_table, cdf_table;
    u32             norm_age_count, uili, acrc, lmcr, crm;
    int             scdi = 0;     /* SCD Index */

    PRINTK("nicstar%d: ns_init_tx_vbr ()\n", card->index);
    XPRINTK ("    pcr= 0x%x, max_pcr= 0x%x, min_pcr= 0x%x\n", 
                        vcc->qos.txtp.pcr, vcc->qos.txtp.max_pcr, vcc->qos.txtp.min_pcr); 
    XPRINTK ("    rif= 0x%x, rdf= Ox%x, cdf= 0x%x\n",
                        vcc->qos.txtp.rif, vcc->qos.txtp.rdf, vcc->qos.txtp.cdf);

    if ((vcc->qos.txtp.max_pcr < vcc->qos.txtp.pcr) || 
                                    (vcc->qos.txtp.pcr < vcc->qos.txtp.min_pcr)) {
        printk("nicstar%d: ABR vc should have max_pcr >= pcr >= min_pcr: %d, %d, %d\n", 
                card->index, vcc->qos.txtp.max_pcr, vcc->qos.txtp.pcr, vcc->qos.txtp.min_pcr);
        return -EINVAL;
    }

    /* get SCD */
    for (scdi = 0; scdi < NS_SCD_NUM; scdi++) {
        if (card->scd2vc[scdi] == NULL) {
            card->scd2vc[scdi] = vc;
            break;
        }
    }
    if (scdi == NS_SCD_NUM) {
        printk("nicstar%d: no SCD available for ABR channel.\n", card->index);
        return -EBUSY;
    }
    vc->adr_scd = card->scd + scdi * NS_SCD_ENTRY_SIZE;

    /* get SCQ */
    scq = get_scq(card, ABR_SCQ, vc->adr_scd);
    if (scq == (scq_info *) NULL) {
        printk("nicstar%d: can't get abr SCQ.\n", card->index);
        card->scd2vc[scdi] = NULL;
        return - ENOMEM;
    }
    vc->scq = scq;

    /* set up SCD */
    u32d[0] = scq->base_pa;
    u32d[1] = (u32) u32d[0]& 0x00001FF0;
    if ((u32d[0]& 0x00001FFF) != u32d[1])
        printk("nicstar: ns_open, error in scq->base address \n");

    u32d[2] = (u32)0xffffffff;
    u32d[3] = (u32)0x00000000;
    ns_write_sram(card, vc->adr_scd, u32d, 4);
    for (j = 1; j < 3; j++)           /* initialize the cache */
    {
        u32d[0] = (u32)0x00000000;
        u32d[1] = (u32)0x00000000;
        u32d[2] = (u32)0x00000000;
        u32d[3] = (u32)0x00000000;
        ns_write_sram(card, vc->adr_scd + 4 *j, u32d, 4);
    }
    XPRINTK("nicstar: ABR SCD written at:0x%x\n", vc->adr_scd);

    /* create ABR entry in the TCT */
    vc->init_er = ns_rate2log (card, vcc->qos.txtp.max_pcr);
    lmcr = (u32) ns_rate2log (card, vcc->qos.txtp.min_pcr);
    vc->lacr = (u32) ns_rate2log (card, vcc->qos.txtp.pcr);
    air_table = 4 + 15 - (vcc->qos.txtp.rif);
    rdf_table = 20 + 15 - (vcc->qos.txtp.rdf);

    if (vcc->qos.txtp.cdf_pres)
            cdf_table = 20 + 15 - (vcc->qos.txtp.cdf);
    else
        cdf_table = rdf_table;

    norm_age_count = 4;
    uili = 0;                                     /* use it or lose it */
    acrc = (u8)(vcc->qos.txtp.adtf_pres);
    crm = 4;

    u32d[0] = NS_TCTE_TYPE_ABR | acrc<<28 | uili<<26 |  vc->adr_scd ;
    u32d[1] = crm << 29;
    u32d[2] = NS_TCTE_TSIF;    
    u32d[3] = (NS_TCTE_HALT | NS_TCTE_IDLE) | (norm_age_count << 21);
    ns_write_sram(card, (conn * 8), u32d, 4);

    u32d[4] = (lmcr << 24);
    u32d[5] = ((cdf_table & 0x3F) << 20) | ((rdf_table & 0x3F) << 14) | 
                                    ((air_table & 0x3F) << 8);
    u32d[6] = 0x00000000;
    u32d[7] = 0x00000000;
    ns_write_sram(card, (conn * 8) + 4, u32d+4, 4);

    return 0;
}


/*  Check VPI and VCI  */

static int ns_check_vc (ns_dev *card, short vpi, int vci)
{
    u16 vp = (u16)vpi;
    u32 vc = (u32)vci;   

    /* vci check */
    if ((vc >> card->vcibits) != (card->vpm & (0xffff >> card->vcibits))) {
        return -EINVAL;
    }
    /* vpi check */
    if ((vp >> card->vpibits) != (card->vpm >> (16 - card->vcibits))) {
        return -EINVAL;
    }
    return 0;
}

static inline int ns_find_conn (ns_dev *card, short vpi, int vci)
{
    return ((vpi & card->maxvpi)<< card->vcibits) | (vci & card->maxvci);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,6,0)
static int ns_open(struct atm_vcc *vcc)
#else
static int ns_open(struct atm_vcc *vcc, short vpi, int vci)
#endif
{
    ns_dev          *card;
    vc_map          *vc;
    int             error;
    int             inuse;          /* tx or rx vc already in use by another vcc */
    u32             conn;
    unsigned char   traffic_class;   

    /* the connection number used to index the TCT */
    card = (ns_dev *) vcc->dev->dev_data;

    PRINTK("nicstar%d: ns_open ()\n", card->index);

    if (vcc->qos.aal != ATM_AAL5 && vcc->qos.aal != ATM_AAL0) {
        printk("nicstar%d: unsupported AAL.\n", card->index);
        return - EINVAL;
    }
#if LINUX_VERSION_CODE < KERNEL_VERSION (2,6,0)
    PRINTK("nicstar%d: opening vpi.vci %d.%d \n", card->index, (int) vpi, vci);
    if ((error = ns_check_vc (card, vpi, vci))) {
        printk ("nicstar%d: vpi.vci %d.%d out of bounds (max vpi= %d, max vci= %d, mask = 0x%x)\n", 
        card->index, (int)vpi, vci, card->maxvpi, card->maxvci, (int)card->vpm);
        return error;
    }

    if ((error = atm_find_ci(vcc, &vpi, &vci))) {
        printk("nicstar%d: error in atm_find_ci().\n", card->index);
        return error;
    }

    conn =  ns_find_conn (card, vpi, vci);
    vcc->vpi = vpi;
    vcc->vci = vci;
#else
    PRINTK("nicstar%d: opening vpi.vci %d.%d \n", card->index, (int) vcc->vpi, vcc->vci);
    conn =  ns_find_conn (card, vcc->vpi, vcc->vci);
#endif

    vc = & (card->vcmap[conn]);
    inuse = 0;
    if (vcc->qos.txtp.traffic_class != ATM_NONE && vc->tx)
        inuse += 1;

    if (vcc->qos.rxtp.traffic_class != ATM_NONE && vc->rx)
        inuse += 2;

    if (inuse) {
        printk("nicstar%d: %s vci already in use.\n", card->index, 
               inuse == 1 ? "tx" : inuse == 2  ? "rx" : "tx and rx");
        return - EINVAL;
    }

    vc->oam_cc = 0;
    vc->closing = 0;
    vc->scq = NULL;
    vcc->dev_data = vc;

    set_bit(ATM_VF_ADDR, &vcc->flags);

    if (vcc->qos.txtp.traffic_class != ATM_NONE) {

        /* NOTE: You are not allowed to modify an open connection's QOS. To change
         * that, remove the ATM_VF_PARTIAL flag checking. There may be other changes
         * needed to do that. 
         */

        if (!test_bit(ATM_VF_PARTIAL, &vcc->flags )) {

            set_bit(ATM_VF_PARTIAL, &vcc->flags);

            traffic_class = vcc->qos.txtp.traffic_class; 

            switch (traffic_class) {

            case ATM_CBR:
                PRINTK("nicstar%d: trying to open CBR vc.\n", card->index);
                error = ns_init_tx_cbr (card, vc, vcc, conn);
                break;

            case (ATM_UBR):     
                PRINTK("nicstar%d: trying to open UBR vc.\n", card->index);
                error = ns_init_tx_ubr (card, vc, vcc, conn);
                break;

            case (ATM_VBR):
                PRINTK("nicstar%d: trying to open VBR vc.\n", card->index);
                error = ns_init_tx_vbr (card, vc, vcc, conn);
                break;

            case (ATM_ABR):
                PRINTK("nicstar%d: trying to open ABR vc.\n", card->index);
                error = ns_init_tx_abr (card, vc, vcc, conn);
                break;

            default:
                printk ("nicstar%d: Unknown traffic class, %d\n", 
                                                card->index, traffic_class);
                return -EPROTONOSUPPORT;

            }   /* switch (traffic_class) */

            if (error) {
                clear_bit(ATM_VF_PARTIAL, &vcc->flags);
                clear_bit(ATM_VF_ADDR, &vcc->flags);
                return error;
            }

            /* start scheduling for non CBR VC*/
            if (vcc->qos.txtp.traffic_class != ATM_CBR) {
                vc->active = 1;
                PRINTK ("nicstar: start scheduling connection %d\n", conn);
                    writel(NS_UPD_INIT_ER | ((u32)vc->init_er) <<16 | conn, card->membase + TCMDQ);
                    writel(NS_START_UPD_LACR | ((u32)vc->lacr) <<16 | conn, card->membase + TCMDQ);

                mdelay (1);
            }
        }
    }

    if (vcc->qos.txtp.traffic_class != ATM_NONE) {
        vc->tx = 1;
        vc->tx_vcc = vcc;
        vc->tbd_count = 0;

    }
    if (vcc->qos.rxtp.traffic_class != ATM_NONE) {
        u32             status;

        vc->rx = 1;
        vc->rx_vcc = vcc;
        vc->rx_iov = NULL;

        /* Open the receive connection */
        if (vcc->qos.aal == ATM_AAL5)
            status = NS_RCTE_AAL5 | NS_RCTE_CONNECTOPEN;
        else
            /* vcc->qos.aal == ATM_AAL0 */
#if defined (NS_RCQ_SUPPORT) && (defined NS_AAL0_TO_RCQ)
            /*USE RAW CELL QUEUE*/
            status = NS_RCTE_RAWCELLINTEN | NS_RCTE_RCQ | NS_RCTE_CONNECTOPEN;
#else
            /* use normal queue */
            status = NS_RCTE_AAL0 | NS_RCTE_CONNECTOPEN;
#endif

#ifdef NS_RCQ_SUPPORT
        status |= NS_RCTE_RAWCELLINTEN;
#endif
        ns_write_sram(card, card->rct + conn * NS_RCT_ENTRY_SIZE, &status, 1);

    }
    set_bit(ATM_VF_READY, &vcc->flags);
#if LINUX_VERSION_CODE < KERNEL_VERSION (2,6,0)
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

/* close connection */
static void ns_close(struct atm_vcc *vcc)
{
    vc_map          *vc = vcc->dev_data;
    ns_dev          *card = vcc->dev->dev_data;
    u32             data;
    int             i;
    int             conn = ns_find_conn (card, vcc->vpi, vcc->vci);
    unsigned long   flags;
    ns_scqe         *scqep;
    scq_info        *scq;
    u32             u32d[4];

    PRINTK("nicstar%d: ns_close (vpi.vci %d.%d)\n", card->index, 
           (int) vcc->vpi, vcc->vci);

    if (!vc) {
        printk ("nicstar%d: trying to close an already closed VC: vpi.vci %d.%d)\n", 
            card->index, (int) vcc->vpi, vcc->vci);
        return;
    }
    ns_grab_res_lock(card, flags);
    if (vc->closing) {
        printk ("nicstar%d: trying to close a closing VC: vpi.vci %d.%d)\n", 
            card->index, (int) vcc->vpi, vcc->vci);
        spin_unlock_irqrestore(&card->res_lock, flags);
        return;
    }
    vc->closing = 1;
    spin_unlock_irqrestore(&card->res_lock, flags);


    clear_bit(ATM_VF_READY, &vcc->flags);
    if (vcc->qos.rxtp.traffic_class != ATM_NONE) {
        u32             addr;
        unsigned long   flags;

        addr = card->rct + conn * NS_RCT_ENTRY_SIZE;
        ns_grab_res_lock(card, flags);
        waitfor_notbusy(card);

        /* stop receive part */
        writel(NS_CMD_CLOSE_CONNECTION | addr << 2, card->membase + CMD);
        spin_unlock_irqrestore(&card->res_lock, flags);
        vc->rx = 0;
        if (vc->rx_iov != NULL) {
            struct sk_buff  *iovb;
            u32             stat;
            u32             fbqp0;
            u32             fbqp1;

            stat = readl(card->membase + STAT);
            fbqp0 = readl(card->membase + FBQP0);
            fbqp1 = readl(card->membase + FBQP1);
            card->bfqc[0] = ns_fbqc_get(fbqp0);
            card->bfqc[1] = ns_fbqc_get(fbqp1);
            PRINTK("nicstar%d: closing a VC with pending rx buffers.\n",
                   card->index);
            iovb = vc->rx_iov;
            recycle_iovec_rx_bufs(card, (struct iovec *) iovb->data,
                                  NS_SKB(iovb)->iovcnt);
            NS_SKB(iovb)->iovcnt = 0;
            ATM_SKB(iovb)->vcc = NULL;
            ns_grab_int_lock(card, flags);
            recycle_iov_buf(card, iovb);
            spin_unlock_irqrestore(&card->int_lock, flags);
            vc->rx_iov = NULL;
        }

        /* Erase RCT entry */
#ifdef NS_RCQ_SUPPORT
        u32d[0] = NS_RCTE_RAWCELLINTEN;
#else 
        u32d[0] = 0x00000000;
#endif 
        u32d[1] = 0x00000000;
        u32d[2] = 0x00000000;
        u32d[3] = 0xFFFFFFFF;
        ns_write_sram(card, addr, u32d, 4);

    }
    if (vcc->qos.txtp.traffic_class != ATM_NONE) {
        vc->tx = 0;

        /* stop transmit part if not scq0 */
        scq = vc->scq;

        if (scq != card->scq0) {
            if (scq != NULL) {
                /* wait until scq is empty */
                do {      
                    ns_grab_scq_lock(card, scq, flags);
                    scqep = scq->next;
                    if (scqep == scq->base)
                        scqep = scq->last;
                    else
                        scqep--;

                    if (scqep == scq->tail) {
                        spin_unlock_irqrestore(&scq->lock, flags);
                        break;                  /* scq is empty */
                    }
                    spin_unlock_irqrestore(&scq->lock, flags);
                    schedule();
                } while (1); 

                if (vcc->qos.txtp.traffic_class == ATM_CBR) {
                    /* Free all TST entries */
                    data = NS_TST_OPCODE_VARIABLE;
                    for (i = 0; i < card->tst_num; i++) {
                        if (card->tste2vc[i] == vc) {
                            ns_write_sram(card, card->tst + i, &data, 1);
                            card->tste2vc[i] = NULL;
                            card->tst_free_entries++;
                        }
                    }
                }
                else if (vcc->qos.txtp.traffic_class != ATM_NONE) {
                    u32 d;
                    d = ns_read_sram (card, conn*8 + 3);
                    if (!(d & NS_TCTE_IDLE)) {
                        writel(NS_HALT | conn, card->membase + TCMDQ);
                    }
                }
                card->scd2vc[ (vc->adr_scd - card->scd) / NS_SCD_ENTRY_SIZE] = NULL;
                free_scq(card, vc->scq, vcc);
                vc->scq = NULL;
            }
        }
        else {
            /* scq0 special process */
            if (vcc->qos.txtp.traffic_class != ATM_NONE) {
               /* Wait for natural end of transmission of all sk_buff 
                  still attached to this connection */
                u32 d;
                u32 transmitting, n=0;

                for (n=0; n < 100; n++) {
                    schedule ();
                    transmitting = 0;
                    ns_grab_scq_lock(card, scq, flags);

                    /* Continuing as long as there are transmit packets for this vcc in scq0 */
                    for (i = 0; i < SCQ_NUM_ENTRIES; i++) {
                        if(scq->skb[i] && ATM_SKB(scq->skb[i])->vcc == vcc) {
                            transmitting = 1;
                            break;
                        }
                    }
                    if (!transmitting) {
                        /* There might be transmit packets for this vcc into the scq0 waiting queue */
                        struct sk_buff *skb = scq->waiting.next;
                        while (skb != (struct sk_buff *)&scq->waiting) {
                            if (ATM_SKB(skb)->vcc == vcc) {
                                transmitting = 1;
                                break;
                            }
                            skb = skb->next;
                        }
                    }
                    spin_unlock_irqrestore(&scq->lock, flags);
                    if (!transmitting)
                        break; 
                } 
                if (transmitting) {
                    /* we should not come here */
                    u32  stat, cfg;
                    stat = readl(card->membase + STAT);
                    cfg = readl(card->membase + CFG);
                    printk("nicstar%d: closing VC %d.%d with pending tx buffers in SCQ0.\n", 
                            card->index,  (int) vcc->vpi, vcc->vci);
                    printk("    STAT = 0x%08X  CFG = 0x%08X  \n", stat, cfg);

                }

                d = ns_read_sram (card, conn*8 + 3);
                if (!(d & NS_TCTE_IDLE)) {
                    writel(NS_HALT | conn, card->membase + TCMDQ);
                }
            }
        }
    }

    vcc->dev_data = NULL;
    clear_bit(ATM_VF_PARTIAL, &vcc->flags);
    clear_bit(ATM_VF_ADDR, &vcc->flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION (2,6,0)
    MOD_DEC_USE_COUNT;
#endif

#ifdef EXTRA_DEBUG
    {
        u32             stat, cfg;
        u32             fbqp0;
        u32             fbqp1;

        stat = readl(card->membase + STAT);
        fbqp0 = readl(card->membase + FBQP0);
        fbqp1 = readl(card->membase + FBQP1);
        cfg = readl(card->membase + CFG);
        printk("    STAT = 0x%08X  CFG = 0x%08X  \n", stat, cfg);
        printk("    TSQ: base = 0x%lx  next = 0x%lx  TSQT = 0x%x \n", 
               (unsigned long) card->tsq.base, (unsigned long) card->tsq.next, readl(card->membase + TSQT));
        printk("    RSQ: base = 0x%lx  next = 0x%lx  last = 0x%lx  RSQT = 0x%x \n", 
               (unsigned long) card->rsq.base, (unsigned long) card->rsq.next, (unsigned long) card->rsq.last,
               readl(card->membase + RSQT));
        printk("    free buffer queue interrupt %s \n", card->fbie ? "enabled" : "disabled");
        printk("    SBCNT = %d  count = %d   LBCNT = %d count = %d \n", 
               ns_fbqc_get(fbqp0), card->bpool[0].count, ns_fbqc_get(fbqp1), card->bpool[1].count);
        printk("    hbpool.count = %d  iovpool.count = %d \n", card->hbpool.count, card->iovpool.count);
    }
#endif 
     PRINTK("nicstar%d: exiting ns_close ()\n", card->index);

}

static void fill_tst(ns_dev *card, int num_entries, vc_map *vc, u32 conn)
{
    u32             new_tst;
    unsigned long   cl;
    int             e, r;
    u32             data;

    /* It would be very complicated to keep the two TSTs synchronized while
       assuring that writes are only made to the inactive TST. So, for now I
       will use only one TST. If problems occur, I will change this again */

    new_tst = card->tst;
    /* Fill procedure */
    for (e = 0; e < card->tst_num; e++) {
        if (card->tste2vc[e] == NULL)
            break;
    }
    if (e == card->tst_num) {
        printk("nicstar%d: No free TST entries found. \n", card->index);
        return;
    }
    r = num_entries;
    cl = card->tst_num;
    data = ns_tste_make(NS_TST_OPCODE_FIXED, conn);
    while (r > 0) {
        if (cl >= card->tst_num && card->tste2vc[e] == NULL) {
            card->tste2vc[e] = vc;
            ns_write_sram(card, new_tst + e, &data, 1);
            XPRINTK("    tst entry: 0x%x, written at 0x%x\n", ns_read_sram(card, new_tst + e), 
                new_tst + e);
            cl -= card->tst_num;
            r--;
        }
        if (++e == card->tst_num) {
            e = 0;
        }
        cl += num_entries;
    }

    /* End of fill procedure */

/*   not usefull since we use only the first TST 
 * 
 *   data = ns_tste_make(NS_TST_OPCODE_END, new_tst<<2);
 *   ns_write_sram(card, new_tst + card->tst_num, &data, 1);
 *   ns_write_sram(card, card->tst_addr + card->tst_num, &data, 1);
 *   card->tst_addr = new_tst;
 */
}

static int ns_send(struct atm_vcc *vcc, struct sk_buff *skb)
{ 
    ns_dev          *card;
    vc_map          *vc;
    scq_info        *scq;
    unsigned long   sflags;
    unsigned long   flags;

    /* TBD flags, not CPU flags */
    card = vcc->dev->dev_data;
    ns_grab_int_lock(card, sflags);

    PRINTK("nicstar%d: ns_send ()\n", card->index);

    if ((vc = (vc_map *) vcc->dev_data) == NULL) {
        printk("nicstar%d: vcc->dev_data == NULL on ns_send().\n", card->index);
        atomic_inc(&vcc->stats->tx_err);
        dev_kfree_skb_any(skb);
        spin_unlock_irqrestore(&card->int_lock, sflags);
        return - EINVAL;
    }
    if (!vc->tx) {
        printk("nicstar%d: Trying to transmit on a non-tx VC.\n", card->index);
        atomic_inc(&vcc->stats->tx_err);
        dev_kfree_skb_any(skb);
        spin_unlock_irqrestore(&card->int_lock, sflags);
        return - EINVAL;
    }
    if (vcc->qos.aal != ATM_AAL5 && vcc->qos.aal != ATM_AAL0) {
        printk("nicstar%d: Only AAL0 and AAL5 are supported.\n", card->index);
        atomic_inc(&vcc->stats->tx_err);
        dev_kfree_skb_any(skb);
        spin_unlock_irqrestore(&card->int_lock, sflags);
        return - EINVAL;
    }
    if (skb_shinfo(skb)->nr_frags != 0) {
        printk("nicstar%d: No scatter-gather yet.\n", card->index);
        atomic_inc(&vcc->stats->tx_err);
        dev_kfree_skb_any(skb);
        spin_unlock_irqrestore(&card->int_lock, sflags);
        return - EINVAL;
    }
    scq = ((vc_map *) vcc->dev_data)->scq;
    ATM_SKB(skb)->vcc = vcc;
    skb_queue_tail(&scq->waiting, skb);
    ns_grab_scq_lock(card, scq, flags);
    ns_push_on_scq (card, scq);
    spin_unlock_irqrestore(&scq->lock, flags);
    spin_unlock_irqrestore(&card->int_lock, sflags);
    return 0;
}

/* try to push a maximum number of skbs on SCQ */

static void ns_push_on_scq (ns_dev *card, scq_info *scq) 
{
    ns_scqe         scqe;
    struct sk_buff  *skb;
    struct atm_vcc  *vcc;
    vc_map          *vc;
    u32             flags;
    u32             buflen;

    PRINTK("nicstar%d: ns_push_on_scq()\n", card->index);

    while ((skb = skb_dequeue(&scq->waiting))) {

        vcc = ATM_SKB(skb)->vcc;
        vc = (vc_map *) vcc->dev_data;
        NS_SKB(skb)->paddr = pci_map_single(card->pcidev, skb->data, skb->len, PCI_DMA_TODEVICE);

        if (skb->nfmark == 0xAB) {
            /* AAL type does not care here */
            ns_rcqe * rawcell = (ns_rcqe*)skb->data;
            PRINTK("nicstar: push OAM cell\n");
            buflen = ATM_CELL_PAYLOAD;
            flags = NS_TBD_OAM | NS_TBD_EOPDU;
            scqe.word_2 = NS_SKB(skb)->paddr + 16;     /* raw cell payload */
            scqe.word_3 = cpu_to_le32(0x00000000);
            scqe.word_4 = cpu_to_le32(rawcell->word_1);       /* header */
        }
        else if (vcc->qos.aal == ATM_AAL5) {
            buflen = (u32)skb->len;
            flags = NS_TBD_AAL5;
            scqe.word_2 = NS_SKB(skb)->paddr;
            scqe.word_3 = cpu_to_le32((u32) skb->len);
            scqe.word_4 = ns_tbd_mkword_4(0, (u32) vcc->vpi, (u32) vcc->vci, (u32)vcc->aal_options, 
                                          ATM_SKB(skb)->atm_options & ATM_ATMOPT_CLP ? 1: 0);
            flags |= NS_TBD_EOPDU;      /* support only one-buffer CS-PDUs */
        }
        else {
             /* (vcc->qos.aal == ATM_AAL0) */
            u32 header =  cpu_to_le32(*((u32 *) skb->data));

#ifdef NS_OAM_SUPPORT
            u32 pti = (header & 0x0E) >> 1;
            u32 vci = (header &  NS_TBD_VCI_MASK) >> NS_TBD_VCI_SHIFT;
  #ifdef NS_BUILD_AAL0_HEADER
            u32 gfc = (header & 0xF0000000) >> 28;
  #else
            u32 vpi = (header & NS_TBD_VPI_MASK) >> NS_TBD_VPI_SHIFT;
  #endif
#else
  #ifdef NS_BUILD_AAL0_HEADER
            u32 gfc = (header & 0xF0000000) >> 28;
            u32 pti = (header & 0x0E) >> 1;
  #else
            u32 vci = (header &  NS_TBD_VCI_MASK) >> NS_TBD_VCI_SHIFT;
            u32 vpi = (header & NS_TBD_VPI_MASK) >> NS_TBD_VPI_SHIFT;
  #endif
#endif

#ifdef NS_OAM_SUPPORT
            if (ns_oam_type(pti, vci)) {
                PRINTK(" cell pushed\n");
                flags = NS_TBD_OAM;
            }
            else
#endif
                flags = NS_TBD_AAL0;
            flags |= NS_TBD_EOPDU;                  /* Payload type 1 - end of pdu */
                                                    /* support only one-buffer CS-PDUs */
            buflen = ATM_CELL_PAYLOAD;              /* i.e., 48 bytes */

            scqe.word_2 = cpu_to_le32(pci_map_single(card->pcidev, skb->data + NS_AAL0_HEADER, 
                                 skb_end_ptr(skb) - skb->data - NS_AAL0_HEADER, PCI_DMA_TODEVICE));

            scqe.word_3 = cpu_to_le32(0x00000000);
#ifdef NS_BUILD_AAL0_HEADER
            scqe.word_4 = ns_tbd_mkword_4(gfc, (u32) vcc->vpi, (u32) vcc->vci, pti,
                                                ATM_SKB(skb)->atm_options & ATM_ATMOPT_CLP ? 1 : 0);
#else
            /* Sanity check */
            if (vpi !=  (u32)vcc->vpi || vci != (u32)vcc->vci ) {
                printk ("nicstar%d: Wrong VPI.VCI, expected %d.%d, forced %d.%d\n", 
                                            card->index, (u32)vcc->vpi, (u32)vcc->vci, vpi, vci);
            }   
            scqe.word_4 = header;
#endif

        }
        flags |= NS_TBD_GTSI | NS_TBD_TSIF;
        scqe.word_1 = ns_tbd_mkword_1(flags, (u32)0, buflen); /* tsrtag is inserted by push_scqe */

        XPRINTK("nicstar%d: ns_push_on_scq(): current scq-base= 0x%lx \n", 
                                                                    card->index, (unsigned long) scq->base);

        /* reenqueue skb and stop if scq is full */
        if (push_scqe(card, vc, scq, &scqe, skb) != 0) {
            XPRINTK("nicstar%d: ns_push_on_scq(): must reenqueue skb since scq is full \n", card->index);
            skb_queue_head(&scq->waiting, skb);    
            pci_unmap_single(card->pcidev, NS_SKB(skb)->paddr, skb->len, PCI_DMA_TODEVICE);
            break;
        }
        atomic_inc(&vcc->stats->tx);
    }
}


static int push_scqe(ns_dev *card, vc_map *vc, scq_info *scq, ns_scqe *tbd, struct sk_buff *skb)
{
    u32             data, i, index;

    PRINTK("nicstar%d: push_scqe ()\n", card->index);

    index = 0x3f & (int)(scq->next - scq->base);
    i = index & 0x1f;

    /* check if enough room for TBD */
    if (vc->scqi[i] != 0) {
        XPRINTK("nicstar%d: no room for TBD\n", card->index);
        return 1;    /* no room for TBD */
    }

    vc->scqi[i] = index | 0x80;

    tbd->word_1 |= i << 20;      /* insert tsrtag into TBD */
    *scq->next = *tbd;          /* copy TBD into SCQ entry */
    scq->skb[index] = skb;

    XPRINTK("Sending skb at 0x%lx (pos %d).\n", (unsigned long) skb, index);
    XPRINTK("TBD: 0x%x, 0x%x, 0x%x, 0x%x\n", tbd->word_1, tbd->word_2, tbd->word_3, tbd->word_4);

    if (scq->next == scq->last)
        scq->next = scq->base;
    else
        scq->next++;

    /* update Tail pointer in the SCD*/
    data = scq->base_pa + sizeof(ns_scqe) * (scq->next - scq->base);       /* scq is consistent memory */
    ns_write_sram(card, scq->scd, &data, 1);

    /* restart connection if idle  */
    if (scq->scq_type != CBR_SCQ && scq != card->scq0) {
        if (!vc->active) {
            writel (NS_START | vc->conn, card->membase + TCMDQ);
            vc->active = 1;
        }
    }
    return 0;
}

static void process_tsq(ns_dev *card)
{
    u32             type;
    scq_info        *scq;
    vc_map          *vc;
    u32             vci, vpi;
    int             conn;
    u32             tsrtag, pos;

    PRINTK("nicstar%d: process_tsq ()\n", card->index);

    if (ns_tsi_isempty(card->tsq.next)){
        XPRINTK("nicstar%d: tsi is empty. returning.\n", card->index);
        return;
    }

    do {
        type = ns_tsi_gettype(card->tsq.next);

        if (type == NS_TSI_TYPE_TBD) {
            /* the TSI is due to a TBD */
            unsigned long   flags;
            vci = ns_tsi_getvci (card->tsq.next);
            vpi = ns_tsi_getvpi (card->tsq.next);
            conn = ns_find_conn (card, vpi, vci);
            vc = &(card->vcmap[conn]);
            scq = vc->scq;
            tsrtag = ns_tsi_gettag(card->tsq.next);
            ns_grab_scq_lock(card, scq, flags);
            pos = 0x3f & vc->scqi[tsrtag];

            drain_scq (card, scq, vc, pos);

            ns_push_on_scq (card, scq);
            spin_unlock_irqrestore(&scq->lock, flags);
        }
        else if (type == NS_TSI_TYPE_IDLE) {
            XPRINTK ("nicstar%d: tsi type is NS_TSI_TYPE_IDLE\n", card->index);
            conn = (ns_tsi_getword_1 (card->tsq.next)) & 0x3fff; 
            if( conn < card->rct_size) { 
                vc = &card->vcmap[conn];
                vc->active = 0;      /* record that connection is no longer active*/
            }
        }
        else {
            XPRINTK ("nicstar%d: unknow tsi type: 0x%x\n", card->index, type);
        }
        ns_tsi_init(card->tsq.next);     /* mark TSI empty */
        card->tsq.next++ ;
        if (card->tsq.next == card->tsq.base + NS_TSQ_NUM_ENTRIES)
            card->tsq.next = card->tsq.base;
    }  
    while (!ns_tsi_isempty(card->tsq.next));

    writel ((u32)((unsigned long)card->tsq.next - 
                        (unsigned long)card->tsq.base), card->membase + TSQH);


}

static void drain_scq(ns_dev *card, scq_info *scq, vc_map *vc, int pos)
{
    struct atm_vcc  *vcc;
    struct sk_buff  *skb;
    int             i;

    PRINTK("nicstar%d: drain_scq (scq= 0x%lx, pos= %d)\n", 
                                        card->index, (unsigned long) scq, pos);
    if (pos >= SCQ_NUM_ENTRIES) {
        printk("nicstar%d: Bad index on drain_scq().\n", card->index);
        return;
    }

    i = (int)(scq->tail - scq->base);
    if (++i == SCQ_NUM_ENTRIES)
        i = 0;

    while (1) {
        vc->scqi[i & 0x1f] = 0;
        skb = scq->skb[i];
        XPRINTK("nicstar%d: freeing skb at 0x%lx (index %d).\n", 
                                    card->index, (unsigned long) skb, i);
        if (skb != NULL) {
            vcc = ATM_SKB(skb)->vcc;

            pci_unmap_single(card->pcidev, NS_SKB(skb)->paddr, skb->len, PCI_DMA_TODEVICE);

            if (vcc == NULL) {
                printk ("nicstar%d: null vcc into sk_buff.\n", card->index);
                dev_kfree_skb_any(skb);
            }
            else {
                if (vcc->pop != NULL)
                    vcc->pop(vcc, skb);
                else
                    dev_kfree_skb_any(skb);
            }
            scq->skb[i] = NULL;
        }
        if (i == pos)
            break;

        if (++i == SCQ_NUM_ENTRIES)
            i = 0;

    } 
    scq->tail = scq->base + pos;
}

static void process_rsq(ns_dev *card)
{
    ns_rsqe         *previous;

    PRINTK("nicstar%d: process_rsq ()\n", card->index);

    if (!ns_rsqe_valid(card->rsq.next))
        return;

    do {
        dequeue_rx(card, card->rsq.next);
        ns_rsqe_init(card->rsq.next);
        previous = card->rsq.next;
        if (card->rsq.next == card->rsq.last)
            card->rsq.next = card->rsq.base;
        else
            card->rsq.next++;
    }
    while (ns_rsqe_valid(card->rsq.next));

    writel((u32)((unsigned long) previous - (unsigned long)card->rsq.base), 
           card->membase + RSQH);
}

static void dequeue_rx(ns_dev *card, ns_rsqe *rsqe)
{
    u32             vpi, vci;
    vc_map          *vc;
    struct sk_buff  *iovb;
    struct iovec    *iov;
    struct atm_vcc  *vcc;
    struct sk_buff  *skb;
    unsigned short  aal5_len;
    int             len;
    u32             fbqp0, fbqp1;
    u32             conn;
    u32             handle;
    u32             buftype, index;

    PRINTK("nicstar%d: dequeue_rx ()\n", card->index);

    fbqp0 = readl(card->membase + FBQP0);
    fbqp1 = readl(card->membase + FBQP1);
    card->bfqc[0] = ns_fbqc_get(fbqp0);
    card->bfqc[1] = ns_fbqc_get(fbqp1);
    handle = le32_to_cpu(rsqe->buffer_handle);
    buftype = (handle >> 16) & 1;
    index = handle & (NS_SKBL_SIZE-1);
    skb = card->fsl[buftype].skb[index];
    card->fsl[buftype].skb[index] = NULL;

    if (skb == NULL) {
        printk ("nicstar%d: error: dequeue null skb\n", card->index) ;
        return;
    }

    pci_unmap_single(card->pcidev, NS_SKB(skb)->paddr, 
                            skb_end_ptr(skb) - skb->data, PCI_DMA_FROMDEVICE);

    vpi = ns_rsqe_vpi(rsqe);
    vci = ns_rsqe_vci(rsqe);
    if (vpi == 0 && vci == 0) {
        printk ("nicstar%d: SDU received on vc 0.0.\n", card->index);
        PRINTK ("    rsqe:\nw1= 0x%x, w2= 0x%x, w3= 0x%x, w4= 0x%x\n", 
                        rsqe->word_1, rsqe->buffer_handle, 
                        rsqe->final_aal5_crc32, rsqe->word_4);

        recycle_rx_buf(card, skb);
        return;
    }
    if (ns_check_vc (card, vpi, vci)) {
        printk("nicstar%d: SDU received on out-of-range vc %d.%d.\n", 
               card->index, vpi, vci);
        recycle_rx_buf(card, skb);
        return;
    }
    conn = ns_find_conn (card, vpi, vci);
    vc = &(card->vcmap[conn]);
    if (!vc->rx) {
        printk("nicstar%d: SDU received on non-rx vc %d.%d.\n", 
                 card->index, vpi, vci);
        recycle_rx_buf(card, skb);
        return;
    }
    vcc = vc->rx_vcc;

    pci_unmap_single(card->pcidev, NS_SKB(skb)->paddr, 
                                    skb_end_ptr(skb) - skb->data, PCI_DMA_FROMDEVICE);
    if (vcc->qos.aal == ATM_AAL0) {
        struct sk_buff  *sb;
        unsigned char   *cell;
        int             i;

        cell = skb->data;
        for (i = ns_rsqe_cellcount(rsqe); i; i--)
        {
            if ((sb = dev_alloc_skb(NS_B0SKBSIZE)) == NULL) {
                printk("nicstar%d: Can't allocate buffers for aal0.\n", card->index);
                atomic_add(i, &vcc->stats->rx_drop);
                break;
            }
            if (!atm_charge(vcc, sb->truesize)) {
                printk("nicstar%d: atm_charge() dropped aal0 packets.\n", card->index);
                atomic_add(i - 1, &vcc->stats->rx_drop);/* already increased by 1 */
                dev_kfree_skb_any(sb);
                break;
            }
            /* Rebuild the header */
            * ((u32 *) sb->data) = le32_to_cpu(rsqe->word_1) << 4 | 
                                   (ns_rsqe_clp(rsqe)? 0x00000001 : 0x00000000);
            if (i == 1 && ns_rsqe_eopdu(rsqe))
                * ((u32 *) sb->data) |= 0x00000002;

            skb_put(sb, NS_AAL0_HEADER);
#ifdef NET_SKBUFF_DATA_USES_OFFSET
            memcpy(sb->head + sb->tail, cell, ATM_CELL_PAYLOAD);
#else
            memcpy(sb->tail, cell, ATM_CELL_PAYLOAD);
#endif
            skb_put(sb, ATM_CELL_PAYLOAD);
            ATM_SKB(sb)->vcc = vcc;
            TIMESTP(sb);
            vcc->push(vcc, sb);
            atomic_inc(&vcc->stats->rx);
            cell += ATM_CELL_PAYLOAD;
        }
        recycle_rx_buf(card, skb);
        return;
    }
    /* To reach this point, the AAL layer can only be AAL5 */

    if ((iovb = vc->rx_iov) == NULL) {
        /* first sk_buff. We must get an iovb to record all SDU buffers */
        iovb = skb_dequeue(&card->iovpool.queue);
        if (iovb == NULL) {
            /* No buffers in the queue */
            iovb = alloc_skb(NS_IOVBUFSIZE, GFP_ATOMIC);
            if (iovb == NULL) {
                printk("nicstar%d: out of iovec buffers.\n", card->index);
                atomic_inc(&vcc->stats->rx_drop);
                recycle_rx_buf(card, skb);
                return;
            }
            NS_SKB(iovb)->buftype = BUF_IOV;
        }
        else if (--card->iovpool.count < card->iovnr.min) {
            struct sk_buff  *new_iovb;

            if ((new_iovb = alloc_skb(NS_IOVBUFSIZE, GFP_ATOMIC)) != NULL) {
                NS_SKB(iovb)->buftype = BUF_IOV;
                skb_queue_tail(&card->iovpool.queue, new_iovb);
                card->iovpool.count++;
            }

       }
        vc->rx_iov = iovb;
        NS_SKB(iovb)->iovcnt = 0;
        iovb->len = 0;
#ifdef NET_SKBUFF_DATA_USES_OFFSET
        iovb->tail=0;
        iovb->data = iovb->head;
#else
        iovb->tail = iovb->data = iovb->head;
#endif
        ATM_SKB(iovb)->vcc = vcc;
        /* IMPORTANT: a pointer to the sk_buff containing the small or large
                            buffer is stored as iovec base, NOT a pointer to the 
                            small or large buffer itself. */
    }
    else if (NS_SKB(iovb)->iovcnt >= NS_MAX_IOVECS) {
        /* Too many skbuff's into this SDU */
        printk("nicstar%d: received too big AAL5 SDU.\n", card->index);
        atomic_inc(&vcc->stats->rx_err);
        recycle_iovec_rx_bufs(card, (struct iovec *) iovb->data, NS_MAX_IOVECS);
        NS_SKB(iovb)->iovcnt = 0;
        iovb->len = 0;
#ifdef NET_SKBUFF_DATA_USES_OFFSET
        iovb->tail = 0;
        iovb->data = iovb->head;
#else
        iovb->tail = iovb->data = iovb->head;
#endif
        ATM_SKB(iovb)->vcc = vcc;
    }
    iov = &((struct iovec *) iovb->data)[NS_SKB(iovb)->iovcnt++];
    iov->iov_base = (void *) skb;
    iov->iov_len = ns_rsqe_cellcount(rsqe) *48;
    iovb->len += iov->iov_len;

    /* Sanity check: we use same chaining method as with the 77211, so the first buffer must be
        a small one from FBQ0, then subsequent buffers, if any, must be large ones from FBQ1 */

    if (NS_SKB(iovb)->iovcnt == 1) {

        if (NS_SKB(skb)->buftype != BUF_B0) {
            printk("nicstar%d: Expected a small buffer, and this is not one.\n", card->index);
            which_list(card, skb);
            atomic_inc(&vcc->stats->rx_err);
            recycle_rx_buf(card, skb);
            vc->rx_iov = NULL;
            recycle_iov_buf(card, iovb);
            return;
        }
    } 
    else {
        /* NS_SKB(iovb)->iovcnt >= 2 */
        if (NS_SKB(skb)->buftype != BUF_B1) {
            printk("nicstar%d: Expected a large buffer, and this is not one.\n", card->index);
            which_list(card, skb);
            atomic_inc(&vcc->stats->rx_err);
            recycle_iovec_rx_bufs(card, (struct iovec *) iovb->data, 
                                    NS_SKB(iovb)->iovcnt);
            vc->rx_iov = NULL;
            recycle_iov_buf(card, iovb);
            return;
        }
    }

    /* Push the SDU to upper layers if we got end of CS-PDU */

    if (ns_rsqe_eopdu(rsqe)) {
        /* This works correctly regardless of the endianness of the host */
        unsigned char *L1L2 = (unsigned char *)((unsigned long)skb->data + iov->iov_len - 6);
        aal5_len = L1L2[0] << 8 | L1L2[1];
        len = (aal5_len == 0x0000) ? 0x10000 : aal5_len;

        /* The actual PDU size is len+8. 
           iovb->len is the cumulated payload of all ATM cells used to carry the PDU.
           iovb->len and len must satisfy this relation:  iovb->len > len+8 > iovb->len - 47    */

        if (ns_rsqe_crcerr(rsqe) || (len + 8 > iovb->len || len + (47 + 8) < iovb->len)) {
            printk("nicstar%d: vcc %d.%d AAL5 PDU error", card->index, vpi, vci);
            if (len + 8 > iovb->len || len + (47 + 8) < iovb->len)
                printk(" - PDU size mismatch: %d %d \n", len+8, iovb->len);
            else
                printk(" - wrong CRC.\n");

            atomic_inc(&vcc->stats->rx_err);
            recycle_iovec_rx_bufs(card, (struct iovec *) iovb->data, 
                                  NS_SKB(iovb)->iovcnt);
            vc->rx_iov = NULL;
            recycle_iov_buf(card, iovb);
            return;
        }

        /* By this point we (hopefully) have a complete AAL5 SDU without errors. */

        XPRINTK ("nicstar%d: AAL5 SDU with %d buffers\n", card->index, 
                                                            NS_SKB(iovb)->iovcnt);

        if (NS_SKB(iovb)->iovcnt == 1) {
            /* The SDU contains one small buffer and skb points to it */
            if (!atm_charge (vcc, skb->truesize)) {
                XPRINTK ("nicstar%d: trash SDU\n", card->index);
                push_rxbufs (card, BUF_B0, skb);
            } 
            else {
                skb_put(skb, len);
                dequeue_sm_buf(card, skb);

#ifdef NS_USE_DESTRUCTORS
                skb->destructor = ns_sb_destructor;
#endif
                ATM_SKB(skb)->vcc = vcc;
                TIMESTP(skb);
                vcc->push(vcc, skb);
                atomic_inc(&vcc->stats->rx);
            }
        }
        else if (NS_SKB(iovb)->iovcnt == 2) {
           /* The SDU contains one small plus one large buffer */

            struct sk_buff  *sb = (struct sk_buff *)(iov - 1)->iov_base;
            /* skb points to the large buffer and sb to the small one */
 
            if (len <= NS_B0BUFSIZE) {
                /* The large buffer is empty */
                if (!atm_charge(vcc, sb->truesize)) {
                    /* free small buffer if upper layers cannot handle it */
                    XPRINTK ("nicstar%d: trash 2-buffer small SDU\n", card->index);
                    push_rxbufs(card, BUF_B0, sb);
                } 
                else {
                    skb_put(sb, len);                
                    dequeue_sm_buf(card, sb);

#ifdef NS_USE_DESTRUCTORS
                    sb->destructor = ns_sb_destructor;
#endif
                    ATM_SKB(sb)->vcc = vcc;
                    TIMESTP(sb);

                    /* transmit the small buffer */
                    vcc->push(vcc, sb);          
                    atomic_inc(&vcc->stats->rx);
                }

                /* free the large buffer */
                push_rxbufs(card, BUF_B1, skb);   
            } 
            else {
                /* len > NS_B0BUFSIZE, the usual case */
                if (!atm_charge(vcc, skb->truesize)) {
                    /* free large buffer if upper layers cannot handle it */
                    XPRINTK ("nicstar%d: trash 2-buffer large SDU\n", card->index);
                    push_rxbufs(card, BUF_B1, skb);
                } 
                else {
                    dequeue_lg_buf(card, skb);

#ifdef NS_USE_DESTRUCTORS
                    skb->destructor = ns_lb_destructor;
#endif

                    /* Put everything into the large buffer */
                    skb_push(skb, NS_B0BUFSIZE);
                    memcpy(skb->data, sb->data, NS_B0BUFSIZE);
                    skb_put(skb, len - NS_B0BUFSIZE);
                    ATM_SKB(skb)->vcc = vcc;
                    TIMESTP(skb);

                    /* transmit the large buffer */
                    vcc->push(vcc, skb);         
                    atomic_inc(&vcc->stats->rx);
                }
                /* free the small buffer */
                push_rxbufs(card, BUF_B0, sb);    
            }
        } 
        else {
            /* Big SDU. Must put everything into a huge buffer */
            struct sk_buff  *hb, *sb, *lb;
            int             remaining, tocopy;
            int             j;

            /* Try to get a huge buffer */
            hb = skb_dequeue(& (card->hbpool.queue));
            if (hb == NULL) {

                /* No buffer available in the queue */
                hb = alloc_skb(NS_HBUFSIZE, GFP_ATOMIC);
                if (hb == NULL) {
                    printk("nicstar%d: Out of huge buffers.\n", card->index);
                    atomic_inc(&vcc->stats->rx_drop);
                    recycle_iovec_rx_bufs(card, (struct iovec *) iovb->data, 
                                            NS_SKB(iovb)->iovcnt);
                    vc->rx_iov = NULL;
                    recycle_iov_buf(card, iovb);
                    return;
                }
                else if (card->hbpool.count < card->hbnr.min) {
                    struct sk_buff  *new_hb;

                    if ((new_hb = alloc_skb(NS_HBUFSIZE, GFP_ATOMIC)) != NULL) {
                        NS_SKB(new_hb)->buftype = BUF_HB;
                        skb_queue_tail(&card->hbpool.queue, new_hb);
                        card->hbpool.count++;
                    }
                }
                NS_SKB(hb)->buftype = BUF_HB;
            }
            else if (--card->hbpool.count < card->hbnr.min) {
                struct sk_buff  *new_hb;

                if ((new_hb = alloc_skb(NS_HBUFSIZE, GFP_ATOMIC)) != NULL) {
                    NS_SKB(new_hb)->buftype = BUF_HB;
                    skb_queue_tail(&card->hbpool.queue, new_hb);
                    card->hbpool.count++;
                }
                if (card->hbpool.count < card->hbnr.min) {
                    if ((new_hb = alloc_skb(NS_HBUFSIZE, GFP_ATOMIC)) != NULL) {
                        NS_SKB(new_hb)->buftype = BUF_HB;
                        skb_queue_tail(&card->hbpool.queue, new_hb);
                        card->hbpool.count++;
                    }
                }
            }

            iov = (struct iovec *) iovb->data;
            if (!atm_charge(vcc, hb->truesize)) {
                /* free all received buffers if upper layers cannot handle this SDU */
                XPRINTK ("nicstar%d: trash huge SDU\n", card->index);
                recycle_iovec_rx_bufs(card, iov, NS_SKB(iovb)->iovcnt);

                /* recycle the huge buffer */
                if (card->hbpool.count < card->hbnr.max) {
                    skb_queue_tail(&card->hbpool.queue, hb);
                    card->hbpool.count++;
                }
                else
                    dev_kfree_skb_any(hb);
            } 
            else {
                /* OK, we got a huge buffer and upper layers can process it */
                /* Copy the small buffer to the huge buffer */
                sb = (struct sk_buff *) iov->iov_base;
                memcpy(hb->data, sb->data, iov->iov_len);
                skb_put(hb, iov->iov_len);
                remaining = len - iov->iov_len;
                iov++;

                /* Free the small buffer */
                push_rxbufs(card, BUF_B0, sb);

                /* Copy all large buffers to the huge buffer and free them */
                for (j = 1; j < NS_SKB(iovb)->iovcnt; j++) {
                    lb = (struct sk_buff *) iov->iov_base;
                    tocopy = MIN(remaining, iov->iov_len);
#ifdef NET_SKBUFF_DATA_USES_OFFSET
                    memcpy(hb->head + hb->tail, lb->data, tocopy);
#else
                    memcpy(hb->tail, lb->data, tocopy);
#endif
                    skb_put(hb, tocopy);
                    iov++;
                    remaining -= tocopy;
                    push_rxbufs(card, BUF_B1, lb);
                }
#ifdef EXTRA_DEBUG
                if (remaining != 0 || hb->len != len)
                    printk("nicstar%d: Huge buffer len mismatch.\n", card->index);

#endif
                ATM_SKB(hb)->vcc = vcc;

#ifdef NS_USE_DESTRUCTORS
                hb->destructor = ns_hb_destructor;
#endif
                TIMESTP(hb);

                /* transmit the huge buffer */
                vcc->push(vcc, hb);
                atomic_inc(&vcc->stats->rx);
            }
        }

        /* The received SDU has been processed, recycle the iovec buffer */
        vc->rx_iov = NULL;
        recycle_iov_buf(card, iovb);
    }
}

#ifdef NS_USE_DESTRUCTORS

static void ns_sb_destructor(struct sk_buff *sb)
{       
    ns_dev          *card;
    u32             fbqp0, fbqp1;

    card = (ns_dev *) ATM_SKB(sb)->vcc->dev->dev_data;
    fbqp0 = readl(card->membase + FBQP0);
    fbqp1 = readl(card->membase + FBQP1);
    card->bfqc[0] = ns_fbqc_get(fbqp0);
    card->bfqc[1] = ns_fbqc_get(fbqp1);
    do
    {
        sb = dev_alloc_skb(NS_B0SKBSIZE);
        if (sb == NULL)
            break;
        skb_queue_tail(&card->bpool[0].queue, sb);
        skb_reserve(sb, NS_AAL0_HEADER);
        push_rxbufs(card, BUF_B0, sb);
    }
    while (card->bfqc[0] < card->bnr[0].min);

}

static void ns_lb_destructor(struct sk_buff *lb)
{       
    ns_dev          *card;
    u32             fbqp0, fbqp1;

    card = (ns_dev *) ATM_SKB(lb)->vcc->dev->dev_data;
    fbqp0 = readl(card->membase + FBQP0);
    fbqp1 = readl(card->membase + FBQP1);
    card->bfqc[0] = ns_fbqc_get(fbqp0);
    card->bfqc[1] = ns_fbqc_get(fbqp1);
    do
    {
        lb = dev_alloc_skb(NS_B1SKBSIZE);
        if (lb == NULL)
            break;

        skb_queue_tail(&card->bpool[1].queue, lb);
        skb_reserve(lb, NS_B0BUFSIZE);
        push_rxbufs(card, BUF_B1, lb);
    }
    while (card->bfqc[1] < card->bnr[1].min);

}

static void ns_hb_destructor(struct sk_buff *hb)
{       
    ns_dev          *card;

    card = (ns_dev *) ATM_SKB(hb)->vcc->dev->dev_data;
    while (card->hbpool.count < card->hbnr.init) {
        hb = alloc_skb(NS_HBUFSIZE, GFP_KERNEL);
        if (hb == NULL)
            break;
        NS_SKB(hb)->buftype = BUF_HB;
        skb_queue_tail(&card->hbpool.queue, hb);
        card->hbpool.count++;
    }
}

#endif /* NS_USE_DESTRUCTORS */

static void recycle_rx_buf(ns_dev *card, struct sk_buff *skb)
{
    if (NS_SKB(skb)->buftype == BUF_B0)
        push_rxbufs(card, BUF_B0, skb);
    else if (NS_SKB(skb)->buftype == BUF_B1)
        push_rxbufs(card, BUF_B1, skb);
    else {
        printk("nicstar%d: What kind of rx buffer is this?\n", card->index);
        dev_kfree_skb_any(skb);
    }
}

static void recycle_iovec_rx_bufs(ns_dev *card, struct iovec *iov, int count)
{       
    for (; count > 0; count--)
        recycle_rx_buf (card, (struct sk_buff *)(iov++)->iov_base);
    
}

static void recycle_iov_buf(ns_dev *card, struct sk_buff *iovb)
{       
    if (card->iovpool.count < card->iovnr.max) {
        skb_queue_tail(&card->iovpool.queue, iovb);
        card->iovpool.count++;
    }
    else
        dev_kfree_skb_any(iovb);
}

static void dequeue_sm_buf(ns_dev *card, struct sk_buff *sb)
{
    SKB_UNLINK(sb, &card->bpool[0].queue);

#ifdef NS_USE_DESTRUCTORS
    if (card->bfqc[0] < card->bnr[0].min)
#else 
    if (card->bfqc[0] < card->bnr[0].init) {
        struct sk_buff  *new_sb;

        if ((new_sb = dev_alloc_skb(NS_B0SKBSIZE)) != NULL) {
            skb_queue_tail(&card->bpool[0].queue, new_sb);
            skb_reserve(new_sb, NS_AAL0_HEADER);
            push_rxbufs(card, BUF_B0, new_sb);
        }
    }
    if (card->bfqc[0] < card->bnr[0].init)
#endif /* NS_USE_DESTRUCTORS */
    {
        struct sk_buff  *new_sb;

        if ((new_sb = dev_alloc_skb(NS_B0SKBSIZE)) != NULL) {
            skb_queue_tail(&card->bpool[0].queue, new_sb);
            skb_reserve(new_sb, NS_AAL0_HEADER);
            push_rxbufs(card, BUF_B0, new_sb);
        }
    }
}

static void dequeue_lg_buf(ns_dev *card, struct sk_buff *lb)
{
    SKB_UNLINK(lb, &card->bpool[1].queue);

#ifdef NS_USE_DESTRUCTORS
    if (card->bfqc[1] < card->bnr[1].min)
#else 
    if (card->bfqc[1] < card->bnr[1].init) {
        struct sk_buff  *new_lb;

        if ((new_lb = dev_alloc_skb(NS_B1SKBSIZE)) != NULL) {
            skb_queue_tail(&card->bpool[1].queue, new_lb);
            skb_reserve(new_lb, NS_B0BUFSIZE);
            push_rxbufs(card, BUF_B1, new_lb);
        }
    }
    if (card->bfqc[1] < card->bnr[1].init)
#endif /* NS_USE_DESTRUCTORS */
    {
        struct sk_buff  *new_lb;

        if ((new_lb = dev_alloc_skb(NS_B1SKBSIZE)) != NULL) {
            skb_queue_tail(&card->bpool[1].queue, new_lb);
            skb_reserve(new_lb, NS_B0BUFSIZE);
            push_rxbufs(card, BUF_B1, new_lb);
        }
    }
}

static int ns_proc_read(struct atm_dev *dev, loff_t *pos, char *page)
{       
    u32             fbqp0, fbqp1;
    ns_dev          *card;
    int             left;

    left = (int) *pos;
    card = (ns_dev *) dev->dev_data;
    fbqp0 = readl(card->membase + FBQP0);
    fbqp1 = readl(card->membase + FBQP1);
    if (!left--)
        return sprintf(page, "Pool   count    min   init    max \n");

    if (!left--)
        return sprintf(page, "Small  %5d  %5d  %5d  %5d \n", 
                       ns_fbqc_get(fbqp0), card->bnr[0].min, card->bnr[0].init, 
                       card->bnr[0].max);

    if (!left--)
        return sprintf(page, "Large  %5d  %5d  %5d  %5d \n", 
                       ns_fbqc_get(fbqp1), card->bnr[1].min, card->bnr[1].init, 
                       card->bnr[1].max);

    if (!left--)
        return sprintf(page, "Huge   %5d  %5d  %5d  %5d \n", card->hbpool.count, 
                       card->hbnr.min, card->hbnr.init, card->hbnr.max);

    if (!left--)
        return sprintf(page, "Iovec  %5d  %5d  %5d  %5d \n", card->iovpool.count, 
                       card->iovnr.min, card->iovnr.init, card->iovnr.max);

    if (!left--) {
        int             retval;

        retval = sprintf(page, "Interrupt counter: %u \n", card->intcnt);
        card->intcnt = 0;
        return retval;
    }
    return 0;
}

static int ns_ioctl(struct atm_dev *dev, unsigned int cmd, void *arg)
{       
    ns_dev          *card;
    pool_levels     pl;
    int             btype;
    unsigned long   flags;

    card = dev->dev_data;
    PRINTK("nicstar%d: ns_ioctl (cmd=0x%x)\n", card->index, cmd);

    switch (cmd)
    {
      case NS_GETPSTAT:
        if (get_user(pl.buftype, & ((pool_levels *) arg)->buftype))
            return - EFAULT;

        switch (pl.buftype)
        {
          case NS_BUFTYPE_SMALL:
            pl.count = ns_fbqc_get(readl(card->membase + FBQP0));
            pl.level.min = card->bnr[0].min;
            pl.level.init = card->bnr[0].init;
            pl.level.max = card->bnr[0].max;
            break;
          case NS_BUFTYPE_LARGE:
            pl.count = ns_fbqc_get(readl(card->membase + FBQP1));
            pl.level.min = card->bnr[1].min;
            pl.level.init = card->bnr[1].init;
            pl.level.max = card->bnr[1].max;
            break;
          case NS_BUFTYPE_HUGE:
            pl.count = card->hbpool.count;
            pl.level.min = card->hbnr.min;
            pl.level.init = card->hbnr.init;
            pl.level.max = card->hbnr.max;
            break;
          case NS_BUFTYPE_IOVEC:
            pl.count = card->iovpool.count;
            pl.level.min = card->iovnr.min;
            pl.level.init = card->iovnr.init;
            pl.level.max = card->iovnr.max;
            break;
          default:
            return - ENOIOCTLCMD;
        }
        if (!copy_to_user((pool_levels *) arg, &pl, sizeof (pl)))
            return (sizeof (pl));
        else
            return - EFAULT;

      case NS_SETBUFLEV:
        if (!capable(CAP_NET_ADMIN))
            return - EPERM;

        if (copy_from_user(&pl, (pool_levels *) arg, sizeof (pl)))
            return - EFAULT;

        if (pl.level.min >= pl.level.init || pl.level.init >= pl.level.max)
            return - EINVAL;

        if (pl.level.min == 0)
            return - EINVAL;

        switch (pl.buftype)
        {
          case NS_BUFTYPE_SMALL:
            if (pl.level.max > TOP_B0)
                return - EINVAL;

            card->bnr[0].min = pl.level.min;
            card->bnr[0].init = pl.level.init;
            card->bnr[0].max = pl.level.max;
            break;
          case NS_BUFTYPE_LARGE:
            if (pl.level.max > TOP_B1)
                return - EINVAL;

            card->bnr[1].min = pl.level.min;
            card->bnr[1].init = pl.level.init;
            card->bnr[1].max = pl.level.max;
            break;
          case NS_BUFTYPE_HUGE:
            if (pl.level.max > TOP_HB)
                return - EINVAL;

            card->hbnr.min = pl.level.min;
            card->hbnr.init = pl.level.init;
            card->hbnr.max = pl.level.max;
            break;
          case NS_BUFTYPE_IOVEC:
            if (pl.level.max > TOP_IOVB)
                return - EINVAL;

            card->iovnr.min = pl.level.min;
            card->iovnr.init = pl.level.init;
            card->iovnr.max = pl.level.max;
            break;
          default:
            return - EINVAL;
        }
        return 0;
      case NS_ADJBUFLEV:
        if (!capable(CAP_NET_ADMIN))
            return - EPERM;

        btype = (unsigned long) arg; 
        switch (btype)
        {
          case NS_BUFTYPE_SMALL:
            while (card->bfqc[0] < card->bnr[0].init) {
                struct sk_buff  *sb;

                sb = dev_alloc_skb(NS_B0SKBSIZE);
                if (sb == NULL)
                    return - ENOMEM;

                skb_queue_tail(&card->bpool[0].queue, sb);
                skb_reserve(sb, NS_AAL0_HEADER);
                push_rxbufs(card, BUF_B0, sb);
            }
            break;
          case NS_BUFTYPE_LARGE:
            while (card->bfqc[1] < card->bnr[1].init) {
                struct sk_buff  *lb;

                lb = dev_alloc_skb(NS_B1SKBSIZE);
                if (lb == NULL)
                    return - ENOMEM;

                skb_queue_tail(&card->bpool[1].queue, lb);
                skb_reserve(lb, NS_B0BUFSIZE);
                push_rxbufs(card, BUF_B1, lb);
            }
            break;
          case NS_BUFTYPE_HUGE:
            while (card->hbpool.count > card->hbnr.init) {
                struct sk_buff  *hb;

                ns_grab_int_lock(card, flags);
                hb = skb_dequeue(&card->hbpool.queue);
                card->hbpool.count--;
                spin_unlock_irqrestore(&card->int_lock, flags);
                if (hb == NULL)
                    printk("nicstar%d: huge buffer count inconsistent.\n", 
                           card->index);
                else
                    dev_kfree_skb_any(hb);

            }
            while (card->hbpool.count < card->hbnr.init) {
                struct sk_buff  *hb;

                hb = alloc_skb(NS_HBUFSIZE, GFP_KERNEL);
                if (hb == NULL)
                    return - ENOMEM;

                NS_SKB(hb)->buftype = BUF_HB;
                ns_grab_int_lock(card, flags);
                skb_queue_tail(&card->hbpool.queue, hb);
                card->hbpool.count++;
                spin_unlock_irqrestore(&card->int_lock, flags);
            }
            break;
          case NS_BUFTYPE_IOVEC:
            while (card->iovpool.count > card->iovnr.init) {
                struct sk_buff  *iovb;

                ns_grab_int_lock(card, flags);
                iovb = skb_dequeue(&card->iovpool.queue);
                card->iovpool.count--;
                spin_unlock_irqrestore(&card->int_lock, flags);
                if (iovb == NULL)
                    printk("nicstar%d: iovec buffer count inconsistent.\n", 
                           card->index);
                else
                    dev_kfree_skb_any(iovb);

            }
            while (card->iovpool.count < card->iovnr.init) {
                struct sk_buff  *iovb;

                iovb = alloc_skb(NS_IOVBUFSIZE, GFP_KERNEL);
                if (iovb == NULL)
                    return - ENOMEM;

                NS_SKB(iovb)->buftype = BUF_IOV;
                ns_grab_int_lock(card, flags);
                skb_queue_tail(&card->iovpool.queue, iovb);
                card->iovpool.count++;
                spin_unlock_irqrestore(&card->int_lock, flags);
            }
            break;
          default:
            return - EINVAL;
        }
        return 0;
      default:
        if (dev->phy && dev->phy->ioctl) {
            return dev->phy->ioctl(dev, cmd, arg);
        }
        else {
            printk("nicstar%d: %s == NULL \n", card->index, 
                   dev->phy
                       ? "dev->phy->ioctl"
                       : "dev->phy");
            return - ENOIOCTLCMD;
        }
    }
}
static void which_list(ns_dev *card, struct sk_buff *skb)
{       
    printk("It's a %s buffer.\n", NS_SKB(skb)->buftype == BUF_B0
               ? "small" : NS_SKB(skb)->buftype == BUF_B1 ? "large"
               :  NS_SKB(skb)->buftype == BUF_HB  ? "huge"
               :  NS_SKB(skb)->buftype == BUF_IOV ? "iovec": "unknown");
}

#ifdef NS_POLLING
static void ns_poll(unsigned long arg)
{       
    int             i;
    ns_dev          *card;
    unsigned long   flags;
    u32             stat_r;

    PRINTK("nicstar: ns_poll ()\n");

    for (i = 0; i < num_cards; i++) {
        card = cards[i];
        if (spin_is_locked(&card->int_lock)) {
            /* Probably it isn't worth spinning */
            continue;
        }
        ns_grab_int_lock(card, flags);
        stat_r = readl(card->membase + STAT);
        if (stat_r & NS_STAT_TSIF) {
            writel(NS_STAT_TSIF, card->membase + STAT);
            printk("nicstar%d: TSI\n", card->index);
            process_tsq(card);
        }
        if (stat_r & NS_STAT_EOPDU) {
            writel(NS_STAT_EOPDU, card->membase + STAT);
            printk("nicstar%d: End of PDU\n", card->index);
            process_rsq(card);
        }
        spin_unlock_irqrestore(&card->int_lock, flags);
    }
    mod_timer(&ns_timer, jiffies + NS_POLL_PERIOD);
}

#endif 

static int ns_parse_mac(char *mac, unsigned char *esi)
{       
    int             i, j;
    short           byte1, byte0;

    if (mac == NULL || esi == NULL)
        return - 1;

    j = 0;
    for (i = 0; i < 6; i++) {
        if ((byte1 = ns_h2i(mac[j++])) < 0)
                     return - 1;

        if ((byte0 = ns_h2i(mac[j++])) < 0)
                     return - 1;

        esi[i] = (unsigned char)(byte1 *16 + byte0);
        if (i < 5) {
            if (mac[j++] != ':')
                return - 1;

        }
    }
    return 0;
}

static short ns_h2i(char c)
{       
    if (c >= '0' && c <= '9')
        return (short)(c - '0');

    if (c >= 'A' && c <= 'F')
        return (short)(c - 'A' + 10);

    if (c >= 'a' && c <= 'f')
        return (short)(c - 'a' + 10);

    return - 1;
}

static void ns_phy_put(struct atm_dev *dev, unsigned char value, 
                       unsigned long addr)
{       
    ns_dev          *card;
    unsigned long   flags;

    card = dev->dev_data;
    ns_grab_res_lock(card, flags);
    waitfor_notbusy(card);
    writel((unsigned long) value, card->membase + DR0);
    writel(NS_CMD_WRITE_UTILITY | 0x00000100 | (addr & 0x000000FF), 
           card->membase + CMD);
    spin_unlock_irqrestore(&card->res_lock, flags);
}

static unsigned char ns_phy_get(struct atm_dev *dev, unsigned long addr)
{       
    ns_dev          *card;
    unsigned long   flags;
    unsigned long   data;

    card = dev->dev_data;
    ns_grab_res_lock(card, flags);
    waitfor_notbusy(card);
    writel(NS_CMD_READ_UTILITY | 0x00000100 | (addr & 0x000000FF), 
           card->membase + CMD);
    waitfor_notbusy(card);
    data = readl(card->membase + DR0)& 0x000000FF;
    spin_unlock_irqrestore(&card->res_lock, flags);
    return (unsigned char) data;
}

static u8
ns_eeprom_byte_rd(ns_dev *card, u32 address)
{       
    int             i, value = 0, command = 3;
    volatile        u32 gp = readl(card->membase + GP) & 0xfffffff8;

    udelay(5);                                         /* make sure idle */
    writel (gp | NS_GP_EECLK | NS_GP_EECS, card->membase + GP);         /* CS and Clock high */
    udelay(5);
    /* toggle in  READ CMD (00000011) */
    for (i=7; i >= 0; i--) {
        writel (gp | ((command >> i)& 1), card->membase + GP);           /* Clock low */
        udelay(5);
        writel (gp | NS_GP_EECLK | ((command >> i)& 1), card->membase + GP);    /* Clock high */
        udelay(5);
    }
    /* toggle in address */
    for (i = 7; i >= 0; i--) {
        writel (gp | ((address >> i)& 1), card->membase + GP);          /* Clock low */
        udelay(5);
        writel (gp | NS_GP_EECLK | ((address >> i)& 1), card->membase + GP);   /* Clock high */
        udelay(5);
    }
    /* read EEPROM data */
    for (i = 7; i >= 0; i--) {
        writel (gp, card->membase + GP);                                  /* Clock low */
        udelay(5);
        value |= (readl(card->membase + GP) & NS_GP_EEDI) >> (16 - i);
        writel (gp | NS_GP_EECLK, card->membase + GP);                     /* Clock high */
        udelay(5);
    }
    writel (gp, card->membase + GP);                                      /* CS and Clock low */
    return ((u8)value);
}

static void
ns_eeprom_rd (ns_dev *card, u32 ee_add, u8 *memptr, u32 len)
{
    while (len-- >0)
        *memptr++ = ns_eeprom_byte_rd(card, ee_add++);
}

#ifdef NS_OAM_SUPPORT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
static void ns_push_oamcell (struct work_struct *work)
{
    struct ns_dev *card =  container_of(work, struct ns_dev, oam_wq);
#else
static void ns_push_oamcell (void *dev_id)
{
    ns_dev *card = (ns_dev *)dev_id;
#endif
    unsigned long   flags;
    unsigned long   sflags;
    ns_rcqe * rawcell;
    vc_map *vc;
    scq_info *scq;
    struct sk_buff  *skb;

    ns_grab_int_lock(card, sflags);
    while ((skb = skb_dequeue(&card->oam_tq))) {
        rawcell = (ns_rcqe*)skb->data;
        vc = &(card->vcmap[rawcell->word_2]);
        scq = vc->scq;
        if (vc->scq == NULL) {
            dev_kfree_skb_any(skb);
            continue;
        }
        skb_queue_tail(&scq->waiting, skb);
        ns_grab_scq_lock(card, scq, flags);
        ns_push_on_scq (card, scq);
        spin_unlock_irqrestore(&scq->lock, flags);        
    }
    spin_unlock_irqrestore(&card->int_lock, sflags);        
}

static void ns_push_on_oam_tq (ns_dev *card, ns_rcqe *rawcell)
{
    struct sk_buff  *skb;
    vc_map *vc = &(card->vcmap[rawcell->word_2]);

    if (!vc->tx) {
        printk ("nicstar%d: Cannot send OAM cell on non-tx VC.\n", card->index);
        return;
    }

    skb = dev_alloc_skb(NS_RCQE_SIZE);
    if (!skb) {
        printk("nicstar%d: Cannot allocate memory for OAM cell.\n", card->index);
        return; 
    }

    skb->truesize = NS_RCQE_SIZE;

    ATM_SKB(skb)->vcc = vc->tx_vcc;
    memcpy(skb->data, (u8*)rawcell, NS_RCQE_SIZE);
    skb_put(skb, NS_RCQE_SIZE);
    skb->nfmark = 0xAB;
    skb_queue_tail(&card->oam_tq, skb);
    schedule_work(&card->oam_wq);      
}

static void ns_activate_deactivate_cc(ns_dev *card, ns_rcqe *rawcell)
{
    u8* pl = rawcell->payload;
    vc_map *vc = &(card->vcmap[rawcell->word_2]);
    u8 oam_dir;

    switch (pl[1] & OAM_MASK_MSGID) {
        case OAM_ACTIVATE_REQUEST:
            printk ("nicstar%d: vcc %d.%d OAM CC activate request\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);          
            oam_dir = pl[1] & OAM_MASK_DIR;
            if (oam_dir)  {
                vc->oam_cc = 1;
                pl[1] = oam_dir | OAM_ACTIVATE_CONFIRMED;                                    
                /* Send the confirmation back */                    
                ns_push_on_oam_tq (card, rawcell);
            }
            else {
                printk ("nicstar%d: vcc %d.%d OAM CC activation with no direction.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);                    
            }
            break;

        case OAM_DEACTIVATE_REQUEST:
            printk ("nicstar%d: vcc %d.%d OAM CC deactivate request.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);
            oam_dir = pl[1] & OAM_MASK_DIR;           
            oam_dir = pl[1] & OAM_MASK_DIR;
            if (oam_dir) {
                vc->oam_cc = 0;
                pl[1] = oam_dir | OAM_DEACTIVATE_CONFIRMED;
                ns_push_on_oam_tq (card, rawcell);
            }
            else {
                printk ("nicstar%d: vcc %d.%d OAM CC activation with no direction.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);                    
            }
            break;
        default:
            PRINTK ("nicstar%d: vcc %d.%d Discard OAM act-deact CC = 0x%02x\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci, pl[1]);
            break;
    }
}


static int ns_process_oam_cell (ns_dev *card, ns_rcqe *rawcell)
{
    u8* pl = rawcell->payload;
    vc_map *vc = &(card->vcmap[rawcell->word_2]);

    PRINTK ("nicstar%d: vcc %d.%d Processing OAM cell\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);

    switch (rawcell->payload[0]) {

        case OAM_ACT_DEACT_CC:
            ns_activate_deactivate_cc(card, rawcell);
            break;
                        
        case OAM_LOOPBACK:
            if (pl[1] & OAM_LOOPBACK_REPLY_MASK) {
                PRINTK ("nicstar%d: vcc %d.%d OAM loopback cell received.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);
 
                /* Reset loopback indication */
                pl[1] = 0;
                if (memcmp (pl+6, ns_all1, 16)) {
                    if (memcmp (pl+6, ns_llid, 16)) {
                        if (memcmp (pl+6, ns_all0, 16)) {
                            printk ("nicstar%d: vcc %d.%d Discard OAM loopback cell with wrong direction.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);
                            break;
                        }
                        else {
                            /* set LLID to all 1s (endpoint)*/
                            memcpy (pl+6, ns_all1, 16);
                        }
                    }
                }
                /* Put reply on transmission */
                ns_push_on_oam_tq (card, rawcell);
            }
            else {
                printk ("nicstar%d: vcc %d.%d Discard OAM loopback cell with wrong direction.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);                
            }
            break;

        case OAM_AIS:
            PRINTK ("nicstar%d: vcc %d.%d OAM AIS received.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);
            /* Don't signal any default */            
            break;
        
        case OAM_RDI:
            PRINTK ("nicstar%d: vcc %d.%d OAM RDI received.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci); 
            /* Must return CC response */
            pl[0] = OAM_CC;        
            ns_push_on_oam_tq (card, rawcell);
            break;

        case OAM_CC:       
            PRINTK ("nicstar%d: vcc %d.%d OAM CC received.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);            
            /* Must return a CC response if CC is enabled on this VC*/
            if (vc->oam_cc) {
                pl[0] = OAM_CC;
                ns_push_on_oam_tq (card, rawcell);
            }
            break;

        default:
            PRINTK ("nicstar%d: vcc %d.%d Unknown OAM cell type.\n", 
                                    card->index, vc->rx_vcc->vpi, vc->rx_vcc->vci);
            return -1;
            break;
    }
    return 0;
        
}
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Christian Bucari <bucari@prosum.net>");
MODULE_DESCRIPTION("NICSTAR2 Driver for PROATM cards");
MODULE_VERSION(NICSTAR2_VERSION);

module_init(nicstar_init);
module_exit(nicstar_cleanup);


